###############################################################################
# IMPORTS
###############################################################################

# DEV  IMPORTS ----------------------------------------------------------------

# python IMPORTS --------------------------------------------------------------
import os
import glob
import re
import inspect

# DS IMPORTS ------------------------------------------------------------------
import numpy  as np
import pandas as pd
from IPython.display import display, Markdown, HTML
import pickle
#---
# DS4N6 IMPORTS ---------------------------------------------------------------

# Common  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
import ds4n6_lib.d4          as d4
import ds4n6_lib.gui         as d4_gui
import ds4n6_lib.unx         as d4_unx
import ds4n6_lib.utils       as d4_utils
# Tools - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
import ds4n6_lib.autoruns    as d4_autoruns
import ds4n6_lib.evtx_parser as d4_evtx_parser
import ds4n6_lib.kansa       as d4_kansa
import ds4n6_lib.kape        as d4_kape
import ds4n6_lib.macrobber   as d4_macrobber
import ds4n6_lib.mactime     as d4_mactime
import ds4n6_lib.plaso       as d4_plaso
import ds4n6_lib.volatility  as d4_volatility
import ds4n6_lib.tshark      as d4_tshark
import ds4n6_lib.sabonis     as d4_sabonis
# Artifacts - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
import ds4n6_lib.evtx        as d4_evtx
import ds4n6_lib.flist       as d4_flist
import ds4n6_lib.fstl        as d4_fstl
import ds4n6_lib.pslist      as d4_pslist
import ds4n6_lib.svclist     as d4_svclist
import ds4n6_lib.winreg      as d4_winreg
import ds4n6_lib.amcache     as d4_amcache
# ML  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import ds4n6_lib.ml     as d4_ml
###############################################################################
# VARIABLES
###############################################################################

# All ds4n6_lib libs
d4libs=['d4_autoruns', 'd4_evtx', 'd4_evtx_parser', 'd4_fstl', 'd4_gui', 'd4_kansa', 'd4_kape', 'd4_plaso', 'd4_unx', 'd4_utils', 'd4_winreg', 'd4_volatility', 'd4_pslist', 'd4_svclist', 'd4_tshark']
# Analysis-oriented ds4n6_libs (with analysis(), etc.)
d4anllibs=['d4_autoruns', 'd4_evtx', 'd4_fstl', 'd4_kansa', 'd4_kape', 'd4_plaso', 'd4_volatility', 'd4_pslist', 'd4_svclist']

###############################################################################
# FUNCTIONS
###############################################################################

# FILE READING FUNCTIONS ######################################################
def read_data_common(evdl, **kwargs):
    """ Read data from files or a folder

        Args: 
            evdl (str): path to file/folder source
            kwargs: read options
        Returns: 
            pandas.Dataframe or dictionary of pandas.DataFrame
    """
    if d4.debug >= 3:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    # Parse Arguments =========================================================

    # Tool Args ---------------------------------------------------------------
    tool                     = kwargs.get('tool',                    '')
    # Path & File Args --------------------------------------------------------
    path_filter              = kwargs.get('path_filter',             '')
    prefix                   = kwargs.get('prefix',                  '')
    suffix                   = kwargs.get('suffix',                  '')
    ext                      = kwargs.get('ext',                     '.csv')
    header_names             = kwargs.get('header_names',            None)
    folder_parsing_mode      = kwargs.get('folder_parsing_mode',     'generic')
    path_prefix              = kwargs.get('path_prefix',             '')
    maxdepth                 = kwargs.get('maxdepth',                4)
    # pluginisdfname -> Each file corresponds to the output of a plugin
    # The D4_plugin field will be populated w/ the file name (minus prefix/ext)
    pluginisdfname            = kwargs.get('pluginisdfname',           False)     
    # Hostname Args -----------------------------------------------------------
    hostname                 = kwargs.get('hostname',                None)
    get_hostname_from_folder = kwargs.get('get_hostname_from_folder', False)
    get_hostname_from_file   = kwargs.get('get_hostname_from_file',   False)
    # Data Args ---------------------------------------------------------------
    separator                = kwargs.get('separator',               ',')
    encoding                 = kwargs.get('encoding',                'utf-8')
    filetype                 = kwargs.get('filetype',                'csv')
    # Harmonization Args ------------------------------------------------------
    do_harmonize             = kwargs.get('harmonize',               True)
    # Output Args -------------------------------------------------------------
    merge_files              = kwargs.get('merge_files'             , False)
    build_all_df             = kwargs.get('build_all_df'            , False)
    # Save Args ---------------------------------------------------------------
    use_pickle               = kwargs.get('use_pickle'              , True)

    if not os.path.exists(evdl):
        print("")
        print("ERROR: File/Folder does not exist.")
        print("         "+evdl)
        return

    # Generic =================================================================
    # Pickle file
    pklrawf = evdl+'.raw.pkl' # Raw
    pklhamf = evdl+'.ham.pkl' # Harmonized

    if do_harmonize:
        pklf = pklhamf
    else:
        pklf = pklrawf
    
    if   tool == "kansa" and folder_parsing_mode != "multiple_hosts_in_single_filetype_folder":
        print("")
        print('WARNING: kansa data reading typically needs the following argument:')
        print('         folder_parsing_mode="multiple_hosts_in_single_filetype_folder"')
        print("")
    elif tool == "kape" and folder_parsing_mode != "single_host_with_categories":
        print("")
        print('WARNING: kape data reading typically needs the following argument:')
        print('         folder_parsing_mode="single_host_with_categories"')
        print("")
       

    #==========================================================================
    # There are 4 parsing modes:
    # - generic
    # - multiple_hosts_one_per_folder
    # - multiple_hosts_in_single_filetype_folder  -> kansa
    # - single_host_with_categories               -> kape
    #
    #==========================================================================
    # generic parsing
    if folder_parsing_mode == 'generic':

        if d4.debug >= 2:
            print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()] folder_parsing_mode -> generic")

        # If pickle exists, read from pickle ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        if os.path.exists(pklf) and use_pickle:
    
            # Read from pickle 
            print("- Saved pickle file found:")
            print("      "+pklf)
            print("- Reading data from pickle file.")
            obj = pickle.load(open(pklf, "rb"))
            print("- Done.")

            return obj
    
        # If pickle does not exist, read original data ~~~~~~~~~~~~~~~~~~~~~~~~
        else:
            # If evdl is a folder -> Read all csv files in that folder
            # If evdl is a file -> Read file
            # Read Directory --------------------------------------------------
            if os.path.isdir(evdl):
                if d4.debug >= 3:
                    print("DEBUG: [com-read_data_common()] [folder_parsing_mode: generic] -> Directory")

                if merge_files:
                    dfall = pd.DataFrame([])
                else:
                    dfs = {}

                    if build_all_df:
                        dfs['all'] = pd.DataFrame([])
                files = d4_utils.find_files_in_folder(evdl, path_filter, ext, maxdepth=maxdepth)

                nfiles2read = len(files)

                if nfiles2read == 0:
                    print('  + WARNING: No files found with extension "' + ext + '"')
                else:
                    print('  + INFO: {} files found to read.\n'.format(nfiles2read))

                    cnt = 1
                    for f in files:
                        fbase = os.path.basename(f)
                        frel  = re.sub("^"+evdl+"/", "", f)

                        if get_hostname_from_folder:
                            dfname = re.sub("/.*", "", frel)
                        else:
                            dfname = fbase
                            if ext != '':
                                dfname = dfname.replace(ext, '')

                        if prefix != '':
                            dfname = re.sub('^'+prefix, '', dfname)

                        if suffix != '':
                            dfname = re.sub(suffix+'$', '', dfname)

                        if get_hostname_from_folder:
                            hostname = dfname
                            kwargs['hostname']=hostname

                        if get_hostname_from_file:
                            hostname = f
                            hostname = re.sub(".*/", "", hostname)
                            hostname = re.sub(r"\.[^\.]*$", "", hostname)

                        print('  + ['+str(cnt)+'/'+str(nfiles2read)+'] %-60s -> %-40s' % (frel, dfname), end='')

                        # Read File - - - - - - - - - - - - - - - - - - - - - - - - - -
                        if os.stat(f).st_size == 0:
                            print("[Empty File]")
                        else:
                            if re.search('.csv$', f, re.IGNORECASE) or re.search('.psv$', f, re.IGNORECASE) or filetype == "csv"  or filetype == "psv" :
                                if encoding == '':
                                    try:
                                        df = pd.read_csv(f, sep=separator, names=header_names)
                                        print('['+str(len(df))+']')
                                    except:
                                        df = pd.DataFrame([])
                                        print("[ERROR]")
                                else:
                                    try:
                                        df = pd.read_csv(f, encoding=encoding, names=header_names, sep=separator)
                                        print('['+str(len(df))+']')
                                    except:
                                        print("[ERROR]")
                                        df = pd.DataFrame([])
                            elif re.search('.evtx', f):
                                df = d4_evtx_parser.read_evtx_file(f)

                            # Fix column names
                            for col in df.columns:
                                # Remove leading / trailing spaces
                                colnew = col.strip()
                                if colnew != col:
                                    df = df.rename(columns={col: colnew})
                            # Harmonize - - - - - - - - - - - - - - - - - - - - - - - - - -
                            dftype = df.d4.df_source_identify()

                            if do_harmonize and len(df) > 0:
                                if pluginisdfname == True:
                                    df = harmonize(df, **kwargs, plugin=dfname)
                                else:
                                    df = harmonize(df, **kwargs)

                                if merge_files:
                                    dfall = pd.concat([dfall, df], ignore_index=True)
                                else:
                                    dfs[dfname] = df

                                    if build_all_df:
                                        dfs['all'] = pd.concat([dfs['all'], dfs[fbase]], ignore_index=True)

                        cnt += 1

                    # Convert D4_Hostname_ col to categorical
                    if merge_files:
                        if len(dfall) != 0:
                            if 'D4_Hostname_' in df.columns:
                                if not dfall['D4_Hostname_'].isna().any():
                                    dfall['D4_Hostname_'] = pd.Categorical(dfall['D4_Hostname_'], categories=dfall['D4_Hostname_'].drop_duplicates())
                    else:
                        if build_all_df:
                            if len(dfall) != 0:
                                if not dfs['all']['D4_Hostname_'].isna().any():
                                    dfs['all']['D4_Hostname_'] = pd.Categorical(dfs['all']['D4_Hostname_'], categories=dfs['all']['D4_Hostname_'].drop_duplicates())


                print("")

                # Save to pickle & return - - - - - - - - - - - - - - - - - - - - - - - 
                if merge_files:
                    save_pickle(evdl, dfall, use_pickle, do_harmonize)
                    return dfall
                else:
                    save_pickle(evdl, dfs, use_pickle, do_harmonize)
                    return dfs

            # Read File -------------------------------------------------------
            elif os.path.isfile(os.path.realpath(evdl)):
                if d4.debug >= 3:
                    print("DEBUG: [com-read_data_common()] [folder_parsing_mode: generic] -> File")

                if tool is not None:
                    print("- Reading " + tool + " file:")
                else:
                    print("- Reading tool file:")
                print("    "+evdl)

                df = pd.read_csv(evdl, encoding=encoding, names=header_names)
                df = df.fillna('')

                print("- No. lines: "+str(len(df)))

                # Harmonize - - - - - - - - - - - - - - - - - - - - - - - - - -
                dftype = df.d4.df_source_identify()
                
                if do_harmonize and len(df) > 0:
                    print("- Harmonizing...")
                    if pluginisdfname == True:
                        df = harmonize(df, **kwargs, plugin=dfname)
                    else:
                        df = harmonize(df, **kwargs)

                    #if dftype == "pandas_dataframe-"+tool+"-raw":
                    #    print("- Harmonizing to "+tool+" HAM...")
                    #    df = harmonize(df, **kwargs)

                # Save to pickle  - - - - - - - - - - - - - - - - - - - - - - - 
                save_pickle(evdl, df, use_pickle, do_harmonize)

                return df

    #==========================================================================
    # Folders contain category for multiple hosts (e.g. kansa) 
    elif folder_parsing_mode == 'multiple_hosts_one_per_folder':

        if d4.debug >= 3:
            print("DEBUG: [com-read_data_common()] folder_parsing_mode -> multiple_hosts_in_single_filetype_folder")


        # If pickle exists, read from pickle ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        if os.path.exists(pklf) and use_pickle:

            # Read from pickle
            print("- Saved pickle file found:")
            print("      "+pklf)
            print("- Reading data from pickle file.")
            obj = pickle.load(open(pklf, "rb"))
            print("- Done.")

            return obj

        # If pickle does not exist, read original data ~~~~~~~~~~~~~~~~~~~~~~~~
        else:
            dfs = {}
            # Let's parse all host files into category-specific dataframes
            dirs = [x[0] for x in os.walk(evdl)]
            for dir in dirs:
                if dir != evdl:
                    catd = dir
                    cat = os.path.basename(catd)
                    dfs[cat] = pd.DataFrame()
                    print('  + Reading csv files for category %-20s into dataframe ->  %-20s' % (cat,cat))
                    hostfs = os.listdir(evdl + "/" + cat)
                    for hostf in hostfs:
                        hostffull = evdl + "/" + cat + '/' + hostf
                        hostname = str.replace(os.path.basename(hostffull), '-' + cat + '.csv', '')
                        try:
                            df = pd.read_csv(hostffull, encoding='utf16', names=header_names)
                        except:                
                            df = pd.DataFrame()    

                        dfs[cat] = pd.concat([dfs[cat], df], ignore_index=True)

                    # Harmonize - - - - - - - - - - - - - - - - - - - - - - - -
                    dftype = dfs[cat].d4.df_source_identify()

                    if do_harmonize  and len(dfs[cat]) > 0:
                        dfs[cat] = harmonize(dfs[cat], **kwargs)

            print("\n\nNOTE: Now you can use the syntax <yourvar>['Category'] to access your dataframe\n")

            # Save to pickle - - - - - - - - - - - - - - - - - - - - - - - - - -
            save_pickle(evdl, dfs, use_pickle, do_harmonize)

            # return  - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
            return dfs

    #==========================================================================
    # Folders contain category for multiple hosts (e.g. kansa) 
    elif folder_parsing_mode == 'multiple_hosts_in_single_filetype_folder':

        if d4.debug >= 3:
            print("DEBUG: [com-read_data_common()] folder_parsing_mode -> multiple_hosts_in_single_filetype_folder")


        # If pickle exists, read from pickle ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        if os.path.exists(pklf) and use_pickle:

            # Read from pickle
            print("- Saved pickle file found:")
            print("      "+pklf)
            print("- Reading data from pickle file.")
            obj = pickle.load(open(pklf, "rb"))
            print("- Done.")

            return obj

        # If pickle does not exist, read original data ~~~~~~~~~~~~~~~~~~~~~~~~
        else:
            dfs = {}
            # Let's parse all host files into category-specific dataframes
            dirs = [x[0] for x in os.walk(evdl)]
            for dir in dirs:
                if dir != evdl:
                    catd = dir
                    cat = os.path.basename(catd)
                    dfs[cat] = pd.DataFrame()
                    print('  + Reading csv files for category %-20s into dataframe ->  %-20s' % (cat,cat))
                    hostfs = os.listdir(evdl + "/" + cat)
                    for hostf in hostfs:
                        hostffull = evdl + "/" + cat + '/' + hostf
                        hostname = str.replace(os.path.basename(hostffull), '-' + cat + '.csv', '')
                        try:
                            df = pd.read_csv(hostffull, encoding='utf16', names=header_names)
                        except:                
                            df = pd.DataFrame()    

                        dfs[cat] = pd.concat([dfs[cat], df], ignore_index=True)

                    # Harmonize - - - - - - - - - - - - - - - - - - - - - - - -
                    dftype = dfs[cat].d4.df_source_identify()

                    if do_harmonize  and len(dfs[cat]) > 0:
                        if pluginisdfname == True:
                            dfs[cat] = harmonize(dfs[cat], **kwargs, plugin=cat)
                        else:
                            dfs[cat] = harmonize(dfs[cat], **kwargs)

            print("\n\nNOTE: Now you can use the syntax <yourvar>['Category'] to access your dataframe\n")

            # Save to pickle - - - - - - - - - - - - - - - - - - - - - - - - - -
            save_pickle(evdl, dfs, use_pickle, do_harmonize)

            # return  - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
            return dfs

    #==========================================================================
    # Folders contain categories for a single host (e.g. kape) 
    elif folder_parsing_mode == 'single_host_with_categories':
    # WARNING: At this point this is very kape-specific.

        if d4.debug >= 3:
            print("DEBUG: [com-read_data_common()] folder_parsing_mode -> single_host_with_categories")

        # If pickle exists, read from pickle ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        if os.path.exists(pklf) and use_pickle:

            # Read from pickle
            print("- Saved pickle file found:")
            print("      "+pklf)
            print("- Reading data from pickle file.")
            obj = pickle.load(open(pklf, "rb"))
            print("- Done.")

            return obj

        # If pickle does not exist, read original data ~~~~~~~~~~~~~~~~~~~~~~~~
        else:
            dfs = {}
            # Let's parse all files into category-specific dataframes
            for rootd, dirs, files in os.walk(evdl):
                filesuniq = set(files)
                for fname in filesuniq:
                    if rootd != evdl:
                        category = re.sub(evdl+'/', '', rootd)
                    else:
                        category = "-"

                    fnamebase, _fnameext = os.path.splitext(fname)
                    if ".csv" in fname:
                        artf = os.path.join(rootd, fname)

                        # Clean-up file name to obtain meaningful artifact description
                        artbase = fnamebase
                        artbase = re.sub('^[0-9][0-9]*_', '', artbase)
                        if path_prefix != "":
                            artbase = re.sub(path_prefix+'_', '', artbase)
                            path_prefix2 = re.sub('__', '_', path_prefix, 1)
                            artbase = re.sub('_'+path_prefix2+'_', '-', artbase)
                            artbase = re.sub('_Output$','', artbase)
                            if category != "-":
                                art = category + '-' + artbase
                            else:
                                art = artbase

                        if not art in dfs:                            
                            dfs[art] = pd.DataFrame()                        
                        else:
                            print("- File repeat: %s" % art)

                        print('  + Reading csv file %-110s into dataframe ->  %-100s \n' % (fname, art),end='')
                        try:
                            df = pd.read_csv(artf, names=header_names)
                        except:
                            df = pd.DataFrame()

                        # Harmonize ---------------------------------------------------------------
                        dftype = df.d4.df_source_identify()
                        if do_harmonize  and len(df) > 0:
                            if pluginisdfname == True:
                                df = harmonize(df, **kwargs, plugin=art)
                            else:
                                df = harmonize(df, **kwargs)
                        
                        dfs[art] = pd.concat([dfs[art], df],ignore_index=True)

            print("\n\nNOTE: Now you can use the syntax <yourvar>['Category'] to access your dataframe\n")

            # Save to pickle - - - - - - - - - - - - - - - - - - - - - - - - - -
            save_pickle(evdl, dfs, use_pickle, do_harmonize)

            # return  - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
            return dfs

    # Unknown parsing type ====================================================
    else:
        print("ERROR: Unknown folder parsing type.")


def save_pickle(evdl, dfs, use_pickle=True, do_harmonize=True):
    """Save a pickle
    Input:
        evdl: evidence path
        dfs:  dataframe to save
        use_pickle : True by default. Bypass value from last function.
        do_harmonize : True by default save as Harmonized, False save as RAW.
    """
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    pklrawf = evdl+'.raw.pkl' # Raw
    pklhamf = evdl+'.ham.pkl' # Harmonized

    if not len(dfs) == 0:
        if use_pickle:
            if do_harmonize:
                if not os.path.exists(pklhamf):
                        # Save to pickle - HAM
                        print("- Saving Harmonized DataFrames dict (dfs) as pickle file:")
                        print("    "+pklhamf) 
                        pickle.dump(dfs, open(pklhamf, "wb"))
                        print("- Done.")
            else:
                if not os.path.exists(pklrawf):
                        # Save to pickle - RAW
                        print("- Saving Raw DataFrames dict (dfs) as pickle file:")
                        print("    "+pklrawf) 
                        pickle.dump(dfs, open(pklrawf, "wb"))
                        print("- Done.")
    
    
# DATA IDENTIFICATION #########################################################

def whatis(obj):
    """Same as data_identify.
    """
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    return data_identify(obj)

def data_identify(obj):
    """Return type of obj.
    Input:
        obj: object to identify
    Return:
        type of object (str, dict, df, or unknown.)
    """    
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    if   isinstance(obj, str):
        return str_source_identify_func(obj)
    elif isinstance(obj, dict):
        return dict_source_identify_func(obj)
    elif isinstance(obj, pd.DataFrame):
        return df_source_identify_func(obj)
    else:
        return "unknown"

def str_source_identify_func(mystr):
    """ Indentify string option in analysis functions

        Args: 
            mystr (str)
        Returns: 
            string type
    """
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    if mystr == "":
        return "str-empty"
    elif mystr == "help":
        return "str-help"
    elif mystr == "list":
        return "str-list"
    else:
        return "str-unknown"

def dict_source_identify_func(dict):
    """ indentify type of source in a dictionary of DF

        Args: 
            dict (dict): dict of DF 
        Returns: 
            type of source

    """
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    # We will use 2 methods in order to identify the dict:
    # - 1: Identify dict by the keys of the dictionary
    # - 2: Identify dict by the cols of the first DF in the dict
    # If method 1 fails then method 2 will be attempted

    # Identify dict by the cols of the first DF in the dict =============

    # NEW METHOD
    eltypes = []
    for key in dict.keys():
        if len(dict[key]) != 0:
            obj = dict[key]
            eltype = data_identify(obj)
            eltypes.append(eltype)

    # Identify the common substring
    eltype1   = eltypes[0]
    eltype1cs = eltype1.split("-")

    if d4.debug >= 3:
        display(eltype1)
        display(eltype1cs)
        display(eltypes)

    etcom = ""
    for et1c in eltype1cs:
        if etcom == "":
            etcom = et1c
        else:
            etcom = etcom + "-" + et1c

        allok = True
        for et in eltypes:
            if not re.match(etcom, et) and allok == True:
                allok = False
                
        if allok == True:
            etcomfinal = etcom

        if d4.debug >= 3:
            print(et1c)
            print(etcom)

    if allok == False:
        etcomfinal = etcomfinal + "-raw"

    if d4.debug >= 3:
        print("FINAL: "+etcomfinal)

    if d4.debug >= 4:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()] eltype: "+eltype)

    if etcomfinal != "unknown":
        return 'dict-'+etcomfinal
    else:
        return "unknown"
    

def df_source_identify_func(df):
    """ Identify the type of a DF

        Args: 
            df (pandas.DataFrame): DF to indentify
        Returns: 
            str: type of DF
    """
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    # autoruns --------------------------------------------------------------------
    autoruns_raw_df_cols = pd.Series(['Time', 'Entry Location', 'Entry', 'Enabled', 'Category', 'Profile', 'Description', 'Signer', 'Company', 'Image Path', 'Version', 'Launch String', 'MD5', 'SHA-1', 'PESHA-1', 'PESHA-256', 'SHA-256', 'IMP'])
    autoruns_ham_df_cols = pd.Series(['EntryLocation', 'Entry', 'Enabled', 'Category', 'Profile', 'Description', 'Signer', 'Company', 'ImagePath', 'Version', 'LaunchString', 'MD5', 'SHA-1', 'PESHA-1', 'PESHA-256', 'SHA-256', 'IMP'])
    # fstl ------------------------------------------------------------------------
    mactime_raw_df_cols = pd.Series(['Date', 'Size', 'Type', 'Mode', 'UID', 'GID', 'Meta', 'File Name'])
    mactime_ham_df_cols = pd.Series(['Size', 'MACB', 'UID', 'GID', 'Meta', 'Deleted_', 'Reallocated_', 'DriveLetter_', 'VSS_', 'EVOName_', 'EvidenceName_', 'Partition_', 'Tag_', 'FilePath_'])
    #^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    dfsrc = "unknown"

    # D4_DataType_-col-based identification ------------------------------------------
    # We will assume that, if the D4_DataType_ col is set, the format is HAM
    if   'D4_DataType_' in df.columns and len(df) > 0 and not pd.isna(df['D4_DataType_'].iloc[0]):
        dfsrc = df['D4_DataType_'].iloc[0] + "-ham"
    # D4_Tool_-col-based identification ----------------------------------------------
    # Tool output can be RAW or HTM
    elif 'D4_Tool_' in df.columns and len(df) > 0 and not pd.isna(df['D4_Tool_'].iloc[0]):
        if df['D4_Plugin_'].iloc[0] == "" or pd.isna(df['D4_Plugin_'].iloc[0]):
            dfsrc = df['D4_Tool_'].iloc[0] + "-raw"
        else:
            dfsrc = df['D4_Tool_'].iloc[0] + "-" + df['D4_Plugin_'].iloc[0] + "-raw"
    # autoruns --------------------------------------------------------------------
    elif   autoruns_raw_df_cols.isin(df.columns).all():
        dfsrc="autoruns-raw"
    elif autoruns_ham_df_cols.isin(df.columns).all():
        dfsrc="autoruns-ham"
    # fstl ------------------------------------------------------------------------
    elif mactime_raw_df_cols.isin(df.columns).all():
        dfsrc="mactime-raw"
    elif mactime_ham_df_cols.isin(df.columns).all():
        dfsrc="mactime-ham"
    # plaso -----------------------------------------------------------------------
    elif 'timestamp_desc' in df.columns and '__container_type__' in df.columns and '__type__' in df.columns and 'data_type' in df.columns:
        dfsrc="plaso"
    # evtx  -----------------------------------------------------------------------
    elif '@xmlns' in df.columns and 'System > Provider > @Name' in df.columns and 'System > TimeCreated > @SystemTime' in df.columns:
        dfsrc="evtx-raw"
    # tshark  ---------------------------------------------------------------------
    elif 'frame.time' in df.columns or 'ip.src' in df.columns:
        dfsrc="tshark"
    # unknown ---------------------------------------------------------------------
    else:
        dfsrc = "unknown"
    objtype = 'pandas_dataframe-' + dfsrc

    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()] objtype: "+objtype)

    return objtype

def get_source_options(source_type):
    source_options = {
        'kape':       d4_kape.get_source_options(),
        'plaso':      d4_plaso.get_source_options(),
        'evtx':       d4_evtx.get_source_options()
    }

    return source_options.get(source_type,[])
    

# simple ======================================================================
def simple_func(df, *args, **kwargs):
    """ Reformat the input df so the data is presented to the analyst in the
        friendliest possible way

    Parameters:
    df  (pd.dataframe):  Input data 
    
    Returns:
    pd.DataFrame: Optionally it will return the filtered dataframe, 
                  only if ret=True is set, constant & hidden columns included
                  If ret_out=True is set, then the output just as it is shown
                  (without constant/hidden columns) will be return
    """
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    # Identify data source
    dfsrc = df.d4.df_source_identify()

    if d4.debug >= 3:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()] df type -> "+dfsrc)

    # TOOLS ================================================
    # autoruns ---------------------------------------------
    if   re.search("^pandas_dataframe-autoruns", dfsrc):
        dfout = df.d4_autoruns.simple(*args, **kwargs)
    # kansa ------------------------------------------------
    elif re.search("pandas_dataframe-kansa", dfsrc):
        dfout = df.d4_kansa.simple(*args, **kwargs)
    # kape  ------------------------------------------------
    elif re.search("pandas_dataframe-kape", dfsrc):
        dfout = df.d4_kape.simple(*args, **kwargs)
    # macrobber ----------------------------------------------
    elif re.search("pandas_dataframe-macrobber-raw", dfsrc):
        dfout = df.d4_macrobber.simple(*args, **kwargs)
    elif re.search("pandas_dataframe-macrobber-ham", dfsrc):
        dfout = df.d4_macrobber.simple(*args, **kwargs)
    # mactime ----------------------------------------------
    elif re.search("pandas_dataframe-mactime-raw", dfsrc):
        dfout = df.d4_mactime.simple(*args, **kwargs)
    elif re.search("pandas_dataframe-mactime-ham", dfsrc):
        dfout = df.d4_mactime.simple(*args, **kwargs)
    # plaso ------------------------------------------------
    elif re.search("^pandas_dataframe-plaso", dfsrc):
        dfout = df.d4_plaso.simple(*args, **kwargs)
    # volatility -------------------------------------------
    elif re.search("^pandas_dataframe-volatility", dfsrc):
        dfout = df.d4_volatility.simple(*args, **kwargs)
    # tshark -----------------------------------------------
    elif re.search("^pandas_dataframe-tshark", dfsrc):
        dfout = df.d4_tshark.simple(*args, **kwargs)
    # sabonis ----------------------------------------------
    elif re.search("^pandas_dataframe-sabonis", dfsrc):
        dfout = df.d4_sabonis.simple(*args, **kwargs)
    # DATA TYPES ===========================================

    # amcache ----------------------------------------------
    elif re.search("pandas_dataframe-amcache-ham$", dfsrc):
        dfout = df.d4_amcache.simple(*args, **kwargs)
    # flist  -----------------------------------------------
    elif re.search("pandas_dataframe-flist-ham$", dfsrc):
        dfout = df.d4_flist.simple(*args, **kwargs)
    # pslist  ----------------------------------------------
    elif re.search("pandas_dataframe-pslist-ham$", dfsrc):
        dfout = df.d4_pslist.simple(*args, **kwargs)
    # svclist  ---------------------------------------------
    elif re.search("pandas_dataframe-svclist-ham$", dfsrc):
        dfout = df.d4_svclist.simple(*args, **kwargs)
    # evtx  ------------------------------------------------
    elif re.search("^pandas_dataframe-evtx-raw$", dfsrc):
        dfout = df.d4_evtx.simple(*args, **kwargs)
    elif re.search("^pandas_dataframe-evtx-ham$", dfsrc):
        dfout = df.d4_evtx.simple(*args, **kwargs)
    # reg --------------------------------------------------
    elif re.search("^pandas_dataframe-winreg_kv-ham$", dfsrc):
        dfout = df.d4_winreg.simple(*args, **kwargs)
    # winservices-------------------------------------------
    elif re.search("^pandas_dataframe-winservices-ham$", dfsrc):
        dfout = df.d4_winservices.simple(*args, **kwargs)
    # unknown ----------------------------------------------
    elif dfsrc == "pandas_dataframe-unknown":
        print(Markdown("<font color='orange'>WARNING: DataFrame source not identified. simple() cannot be aplied.</font>"))
        dfout = df
    else:
        print("ERROR: [com] [simple_func] Unexpected error. Unhandled df source ("+dfsrc+").")
        dfout = df

    return dfout

def simple_common_series(se, *args, **kwargs):
    """ Reformat the input Series so the data is presented to the analyst in the
        friendliest possible way

    Parameters:
    se  (pd.Series):  Input data 
    
    """
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    display(se.value_counts().to_frame())
    if se.dtype == 'int64' or se.dtype == 'float64':
        display("Max Value: {}".format(se.max()))
        display("Min Value: {}".format(se.min()))

def simple_common(df, *args, **kwargs):
    """ Reformat the input df so the data is presented to the analyst in the
        friendliest possible way

    Parameters:
    df  (pd.dataframe):  Input data 
    
    Returns:
    pd.DataFrame: Optionally it will return the filtered dataframe, 
                  only if ret=True is set, constant & hidden columns included
                  If ret_out=True is set, then the output just as it is shown
                  (without constant/hidden columns) will be return
    """
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    # Arg. Parsing -------------------------------------------------------------
    hiddencolsuser         = kwargs.get('hiddencolsuser',         []) # WARNING: This arg must be specified as list: ['xxx']
    hiddencols             = kwargs.get('hiddencols',             []) # WARNING: This arg must be specified as list: ['xxx']
    nonhiddencols          = kwargs.get('nonhiddencols',          []) # WARNING: This arg must be specified as list: ['xxx']
    maxdfbprintlines       = kwargs.get('maxdfbprintlines',       20)
    max_rows               = kwargs.get('max_rows',               maxdfbprintlines)
    collapse_constant_cols = kwargs.get('collapse_constant_cols', True)
    hide_cols              = kwargs.get('hide_cols',              True)
    wrap_cols              = kwargs.get('wrap_cols',              True)
    round_ts_to_secs       = kwargs.get('round_ts_to_secs',       True)
    beautify_cols          = kwargs.get('beautify_cols',          True)


    # Show whatever selected on the screen or don't show anything   
    out_opt                = kwargs.get('out',                    True)
    # Show the resulting DF on the screen
    out_df_opt             = kwargs.get('out_df',                 True)
    # Show the resulting Metadata (Constant/Hidden cols, etc.) on the screen
    out_meta_opt           = kwargs.get('out_meta',               True)
    # Return the resulting DataFrame
    ret_opt                = kwargs.get('ret',                    False)
    # Return the resulting DataFrame as shown in the screen
    # (without hidden cols, collapsed cols, etc.; not in full)
    ret_out_opt            = kwargs.get('ret_out',                False)

    # Check if out is set (will be needed later)
    outargset = 'out' in kwargs.keys()

    # Initialize vars
    out      = out_opt
    out_df   = out_df_opt
    out_meta = out_meta_opt
    ret      = ret_opt
    ret_out  = ret_out_opt

    hiddencolsdf = pd.DataFrame([])

    # ret implies no out, unless out is set specifically
    if ('ret' in kwargs.keys() or 'ret_out' in kwargs.keys()) and not outargset and (ret or ret_out):
        # If ret is set to True by the user we will not provide stdout output,
        # unless the user has specifically set out = True
        out      = False # Do not output to the screen at all
        out_df   = False # Do not output to the screen the DF, only the headers
        out_meta = False # Do not output to the screen the Metadata

    if 'out' in kwargs.keys():
        if out_opt == False:
            out_df   = False
            out_meta = False

    if 'out_df' in kwargs.keys():
        if out_df_opt:
            out    = True
            out_df = True

    if 'out_meta' in kwargs.keys():
        if out_meta_opt:
            out      = True
            out_meta = True

    if 'ret_out' in kwargs.keys():
        if ret_out:
            ret = True

    # Health Check -------------------------------------------------------------
    dfnrows = len(df)

    if dfnrows == 0:
        print("ERROR: Empty DataFrame.")
        return

    # Artifact / Tool specific processing --------------------------------------
    #
    # <INSERT HERE THE ARTIFACT / TOOL SPECIFIC PROCESSING ACTIONS>
    #

    # Var. Init. ---------------------------------------------------------------
    dfb = df

    # Hidden Columns - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    if hide_cols:
        if d4.debug >= 3:
            print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()] hide_cols")

        for col in hiddencols:
            if col not in nonhiddencols:
                if col in dfb.columns:
                    hiddencolsdf = pd.concat([hiddencolsdf, pd.DataFrame([col])], ignore_index=True)
                    dfb = dfb.drop(columns=[col])

        for col in hiddencolsuser:
            if col not in nonhiddencols:
                if col in dfb.columns:
                    hiddencolsdf = pd.concat([hiddencolsdf, pd.DataFrame([col])], ignore_index=True)
                    dfb = dfb.drop(columns=[col])

    # Round TStamp columns to seconds  - - - - - - - - - - - - - - - - - - - - -
    if round_ts_to_secs:
        for col in dfb.select_dtypes(include=[np.datetime64]).columns:
            dfb[col] = dfb[col].dt.ceil("S")

    # Constant Columns - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    if collapse_constant_cols:
        if d4.debug >= 3:
            print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()] collapse_constant_cols")

        concolsdf, dfb = d4_utils.collapse_constant_columns(dfb)

    # Wrap wide columns  - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    if wrap_cols:
        if d4.debug >= 3:
            print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()] wrap_cols")

        for col in dfb.columns:
            if isinstance(df[col].iloc[0], str):
                df[col] = df[col].str.wrap(100)
                beautify_cols = True

    # Beautify columns - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    if beautify_cols and out_df and len(dfb) >= max_rows:
        if d4.debug >= 3:
            print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()] beautify_cols")

        print('WARNING: Too many rows (>'+str(max_rows)+') in DataFrame for beautification. Returning non-beautified output.')
        print('         Specify the "max_rows=" option to increase this value')
        beautify_cols = False
        pd.set_option("max_rows", max_rows)

    # DISPLAY ==========================================================
    nhiddencols = len(hiddencolsdf)

    if out:
        if out_meta:
            print("")
            display(Markdown("**Statistics:**\n<br>No. Entries: "+str(dfnrows)))

            if collapse_constant_cols:
                show_constant_cols = True
            else:
                show_constant_cols = False

            if hide_cols and nhiddencols != 0:
                show_hidden_cols = True
            else:
                show_hidden_cols = False

            if show_constant_cols and show_hidden_cols:
                d4_utils.display_side_by_side([hiddencolsdf, concolsdf], ['HIDDEN COLUMNS', 'CONSTANT COLUMNS'])
            elif show_constant_cols and show_hidden_cols == False:
                display(Markdown("**Constant Columns**"))
                max_rows = pd.get_option("display.max_rows")
                pd.set_option("display.max_rows", None)
                display(concolsdf)
                pd.set_option("display.max_rows", max_rows)
            elif show_constant_cols == False and show_hidden_cols:
                display(Markdown("**Hidden Columns**"))
                display(hiddencolsdf)

        if out_df:
            if beautify_cols:
                if dfnrows == 1:
                    display(dfb.T.reset_index().style.set_properties(**{'text-align': 'left', 'white-space': 'pre-wrap' }))
                else:
                    display(dfb.reset_index().style.set_properties(**{'text-align': 'left', 'white-space': 'pre-wrap' }))
            else:
                if dfnrows == 1:
                    display(dfb.T)
                else:
                    display(dfb)

    # RETURN  ==========================================================
    if ret:
        if ret_out:
            return dfb
        else:
            return df

# harmonize ===================================================================
def harmonize(*args, **kwargs):
    """ Alias for harmonize_func
    """
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    return harmonize_func(*args, **kwargs) 

def ham(*args, **kwargs):
    """ Alias for harmonize_func
    """
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    return harmonize_func(*args, **kwargs) 

def harmonize_func(*args, **kwargs):
    """ Convert DF in HAM format

        Args: 
            df (pandas.DataFrame): DF to harmonize
            kwargs(dict): harmonize options
        Returns: 
            pandas.DataFrame in HAM Format
    """
    def syntax():
        print('Syntax: harmonize(obj)\n')

    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    if len(args) == 0:
        syntax()
        return

    obj = args[0]

    tool         = kwargs.get('tool', None)

    # Identify data source
    objtype = data_identify(obj)

    # If the orchestrator or tool arguments are set, we directly invoke the 
    # corresponding module...
    if not tool == None:
        return eval('d4_'+tool).harmonize(*args, **kwargs)

    # ... otherwise, we will try to identify what type of data it is
    # and then invoke the corresponding module

    # help - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    if   objtype == "str-help":
        syntax()
    # atrs - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    elif re.search("^dict-pandas_dataframe-autoruns", objtype) or re.search("^pandas_dataframe-autoruns", objtype):
        return d4_autoruns.harmonize(*args, **kwargs)
    # fstl - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    elif re.search("^dict-pandas_dataframe-fstl", objtype) or re.search("^pandas_dataframe-fstl", objtype):
        return obj
    # kansa  - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    elif re.search("^dict-pandas_dataframe-kansa", objtype) or re.search("^pandas_dataframe-kansa", objtype):
        return d4_kansa.harmonize(*args, **kwargs)
    # kape - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    elif re.search("^dict-pandas_dataframe-kape", objtype) or re.search("^pandas_dataframe-kape", objtype):
        return d4_kape.harmonize(*args, **kwargs)
    # evtx - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    elif re.search("^dict-pandas_dataframe-evtx", objtype) or re.search("^pandas_dataframe-evtx", objtype):
        return d4_evtx.harmonize(*args, **kwargs)
    # fstl - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    elif re.search("^dict-pandas_dataframe-mactime", objtype) or re.search("^pandas_dataframe-mactime", objtype):
        return obj
    # plaso  - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    elif re.search("^dict-pandas_dataframe-plaso", objtype) or re.search("^pandas_dataframe-plaso", objtype):
        return d4_plaso.harmonize(*args, **kwargs)
    # tshark  - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    elif re.search("^dict-pandas_dataframe-tshark", objtype) or re.search("^pandas_dataframe-tshark", objtype):
        return d4_tshark.harmonize(*args, **kwargs)
    # sabonis - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    elif re.search("^dict-pandas_dataframe-sabonis", objtype) or re.search("^pandas_dataframe-sabonis", objtype):
        return d4_sabonis.harmonize(*args, **kwargs)
    # volatility - - - - - - - - - - - - - - - - - - - - - - - - - 
    elif re.search("^dict-pandas_dataframe-volatility", objtype) or re.search("^pandas_dataframe-volatility", objtype):
        return d4_volatility.harmonize(*args, **kwargs)
    elif objtype == "unknown":
        #display(Markdown("<font color='orange'>WARNING: [d4com] Object source not identified. harmonize() cannot be aplied.</font>"))
        return obj
    else:
        #display(Markdown("<font color='orange'>WARNING: [d4com] Object source ("+objtype+") not supported for harmonization.</font>"))
        return obj

def harmonize_common(df, **kwargs):
    """ Convert DF in HAM format, common operations for all kind of DF

        Args: 
            df (pandas.DataFrame): DF to harmonize
            kwargs(dict): harmonize options
        Returns: 
            pandas.DataFrame in HAM Format
    """
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    orchestrator      = kwargs.get('orchestrator',      None)
    tool              = kwargs.get('tool',              None)
    plugin            = kwargs.get('plugin',            None)
    datatype          = kwargs.get('datatype',          None)
    hostname          = kwargs.get('hostname',          None)
    drop_non_ham_cols = kwargs.get('drop_non_ham_cols', False)
    convert_string    = kwargs.get('convert_string',    True)

    # If we only want HAM cols, let's drop the others asap for performance
    if drop_non_ham_cols:
        df.drop(columns=df.columns[~df.columns.str.contains('_$')],inplace=True)

    # Change <NA> to NaN
    # This needs to be done before the "covert_dtypes" because otherwise
    # convert_dtypes will not properly identify timestamp columns
    from pandas.api.types import is_string_dtype

    # Try to adjust timestamp cols to datetime
    for col in df.columns:
        if df[col].dtypes == object:
            if type(df[col].iloc[0]) == str:
                df[col] = pd.to_datetime(df[col], errors='ignore')
    #df = df.apply(lambda col: pd.to_datetime(col, errors='ignore') 
    #              if col.dtypes == object
    #              else col, 
    #              axis=0)

    # Change dtypes via auto-detect 
    # (Note: convert_dtypes does not convert timestamp cols to datetime,
    #        that's why we did it before above)
    if not df.empty:
        df = df.convert_dtypes(convert_string=convert_string)

    # Adjust column names
#    for col in df.columns:
        # Remove spaces from column names
#        colnew = re.sub(" ", "_", col)
#        df = df.rename(columns={col: colnew})

    # These settings may be already set if the data has been generated by 
    # a different collection tools (e.g. kansa -> autoruns) and read/parsed
    # by a different d4 library
    # Or some of these columns may have been created by the pre-processing 
    # section of the calling d4 harmonize function
    if not 'D4_Hostname_' in df.columns:
        df.insert(0, 'D4_Hostname_', hostname)
    if not 'D4_Plugin_' in df.columns:
        df.insert(0, 'D4_Plugin_', plugin)
    if not 'D4_Tool_' in df.columns:
        df.insert(0, 'D4_Tool_', tool)
    if not 'D4_Orchestrator_' in df.columns:
        df.insert(0, 'D4_Orchestrator_', orchestrator)
    if not 'D4_DataType_' in df.columns:
        df.insert(0, 'D4_DataType_', datatype)

    # Resort columns
    cols = df.columns.tolist()
    cols.insert(0, cols.pop(cols.index('D4_Hostname_')))
    cols.insert(0, cols.pop(cols.index('D4_Plugin_')))
    cols.insert(0, cols.pop(cols.index('D4_Tool_')))
    cols.insert(0, cols.pop(cols.index('D4_Orchestrator_')))
    cols.insert(0, cols.pop(cols.index('D4_DataType_')))
    df = df.reindex(columns=cols)

    # Set D4_* columns to categorical
    d4_cols = ['D4_Hostname_', 'D4_Plugin_', 'D4_Tool_', 'D4_Orchestrator_', 'D4_DataType_']

    for col in d4_cols:
        if len(df[col].drop_duplicates()) == 1:
            if df[col].isna().any():
                df[col] = pd.Categorical(df[col], categories=None)
            else:
                df[col] = pd.Categorical(df[col], categories=df[col].drop_duplicates())

    return df

# analysis ====================================================================
def analysis(*args, **kwargs):
    """ Redirects execution to analysis_func()
    """
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    return analysis_func(*args, **kwargs) 

def anl(*args, **kwargs):
    """ Redirects execution to analysis_func()
    """
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    return analysis_func(*args, **kwargs) 

def analysis_func(*args, **kwargs):
    """ Umbrella function that redirects to different types of analysis 
        available on the input data

    Parameters:
    obj:          Input data (typically DF or dict of DFs)
    
    Returns:
    pd.DataFrame: Refer to each specific analysis function
    """
    def syntax():
        print('Syntax: analysis(obj, "analysis_type")\n')
        d4list("list")

    def d4list(objtype):
        for d4lib in d4anllibs:
            mystr=d4lib+'.analysis("list")'
            exec(mystr)
            print("")

    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    if len(args) == 0:
        syntax()
        return

    obj = args[0]

    # Identify data source
    objtype = data_identify(obj)

    # help - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    if   objtype == "str-help":
        syntax()
    # Tools ------------------------------------------------------ 
    # autoruns - - - - - - - - - - - - - - - - - - - - - - - - - - 
    elif re.search("^dict-pandas_dataframe-autoruns", objtype) or re.search("^pandas_dataframe-autoruns", objtype):
        return d4_autoruns.analysis(*args, **kwargs)
    # fstl - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    elif re.search("^dict-pandas_dataframe-flist", objtype) or re.search("^pandas_dataframe-flist", objtype):
        return d4_flist.analysis(*args, **kwargs)
    # kansa  - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    elif re.search("^dict-pandas_dataframe-kansa", objtype) or re.search("^pandas_dataframe-kansa", objtype):
        return d4_kansa.analysis(*args, **kwargs)
    # kape - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    elif re.search("^dict-pandas_dataframe-kape", objtype) or re.search("^pandas_dataframe-kape", objtype):
        return d4_kape.analysis(*args, **kwargs)
    # mactime  - - - - - - - - - - - - - - - - - - - - - - - - - - 
    elif re.search("^dict-pandas_dataframe-mactime", objtype) or re.search("^pandas_dataframe-mactime", objtype):
        return d4_mactime.analysis(*args, **kwargs)
    # plaso  - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    elif re.search("^dict-pandas_dataframe-plaso", objtype) or re.search("^pandas_dataframe-plaso", objtype):
        return d4_plaso.analysis(*args, **kwargs)
    # volatility - - - - - - - - - - - - - - - - - - - - - - - - - 
    elif re.search("^dict-pandas_dataframe-volatility", objtype) or re.search("^pandas_dataframe-volatility", objtype):
        return d4_volatility.analysis(*args, **kwargs)
    # tshark - - - - - - - - - - - - - - - - - - - - - - - - - 
    elif re.search("^dict-pandas_dataframe-tshark", objtype) or re.search("^pandas_dataframe-tshark", objtype):
        return d4_tshark.analysis(*args, **kwargs)
    # Artifacts -------------------------------------------------- 
    elif re.search("^dict-pandas_dataframe-evtx", objtype) or re.search("^pandas_dataframe-evtx", objtype):
        return d4_evtx.analysis(*args, **kwargs)
    elif re.search("^pandas_dataframe-pslist-ham", objtype):
        return d4_pslist.analysis(*args, **kwargs)
    elif objtype == "unknown":
        display(Markdown("<font color='orange'>WARNING: [d4com] Object source not identified. analysis() cannot be aplied.</font>"))
        # return obj
        return 
    else:
        display(Markdown("<font color='orange'>WARNING: [d4com] Object source ("+objtype+") not supported for analysis.</font>"))
        # return obj
        return

def find_anomalies(df, **kwargs):
    return d4_ml.find_anomalies(df, **kwargs)


# ACCESSOR ####################################################################
@pd.api.extensions.register_dataframe_accessor("d4")
class Ds4n6commonAccessor:
    def __init__(self, pandas_obj):
            self._obj = pandas_obj

    def df_source_identify(self):
        df=self._obj
        return df_source_identify_func(df)

    def simple(self, *args, **kwargs):
        df=self._obj
        return simple_func(df, *args, **kwargs)

    def anl(self):
        df=self._obj
        return analysis_func(df)

@pd.api.extensions.register_series_accessor("simple")
class Ds4n6SeriesCommonSimpleAccessor:
    def __init__(self, pandas_obj):
            self._obj = pandas_obj

    def __call__(self, *args, **kwargs):
        se=self._obj
        return simple_common_series(se, *args, **kwargs)

@pd.api.extensions.register_dataframe_accessor("simple")
class Ds4n6CommonSimpleAccessor:
    def __init__(self, pandas_obj):
            self._obj = pandas_obj

    def __call__(self, *args, **kwargs):
        df=self._obj
        return simple_func(df, *args, **kwargs)

# spl() -> Short alias for simple()
@pd.api.extensions.register_dataframe_accessor("spl")
class Ds4n6CommonSplAccessor:
    def __init__(self, pandas_obj):
            self._obj = pandas_obj

    def __call__(self, *args, **kwargs):
        df=self._obj
        return simple_func(df, *args, **kwargs)

@pd.api.extensions.register_dataframe_accessor("analysis")
class Ds4n6CommonAnalysisAccessor:
    def __init__(self, pandas_obj):
            self._obj = pandas_obj

    def __call__(self, *args, **kwargs):
        df=self._obj
        return analysis_func(df, *args, **kwargs)

# anl() -> Short alias for analysis()
@pd.api.extensions.register_dataframe_accessor("anl")
class Ds4n6CommonAnlAccessor:
    def __init__(self, pandas_obj):
            self._obj = pandas_obj

    def __call__(self, *args, **kwargs):
        df=self._obj
        return analysis_func(df, *args, **kwargs)

@pd.api.extensions.register_dataframe_accessor("xgrep")
class Ds4n6CommonXgrepAccessor:
    def __init__(self, pandas_obj):
            self._obj = pandas_obj

    def __call__(self, *args, **kwargs):
        return d4_unx.xgrep_func(self._obj, *args, **kwargs)



# DS4N6
#
# Description: Library of functions to apply Data Science to forensics artifacts
#

###############################################################################
# INFO
###############################################################################
# Recommended "import as": d4evtx_parser

###############################################################################
# IMPORTS
###############################################################################

# DEV  IMPORTS ----------------------------------------------------------------

# python IMPORTS --------------------------------------------------------------
import os
import glob
import re
import time
import pickle
import inspect
from tqdm import tqdm

# DS IMPORTS ------------------------------------------------------------------
import numpy  as np
import pandas as pd
import matplotlib.pyplot as plt
import Evtx.Evtx as evtx
import Evtx.Views as e_views

# DS4N6 IMPORTS ---------------------------------------------------------------
import ds4n6_lib.d4     as d4
import ds4n6_lib.common as d4com
import ds4n6_lib.gui    as d4gui
import ds4n6_lib.utils  as d4utl
import ds4n6_lib.unx    as d4unx
import ds4n6_lib.evtx   as d4evtx

###############################################################################
# FUNCTIONS
###############################################################################

# .evtx READING FUNCTIONS #####################################################

# Read evtx File(s) / Directory ===============================================

def evtx_xml(evtxf):
    """ Converts evtx file to xml string
    Input:
        evtxf: evtx filename.
    Return:
        return the filename as xml string.
    """  

    thistr = ''
    with evtx.Evtx(evtxf) as log:
        thistr = e_views.XML_HEADER
        thistr = thistr + '<Events>'
        for record in tqdm(log.records()):
            try:
                thistr = thistr + record.xml()
            except:
                # key error 138 in Microsoft-Windows-Ntfs%4Operational.evtx file.
                #print("XML Error: %s" % e)
                pass
        thistr = thistr + '</Events>'

    return thistr

def run_tool(path,verbose=True, **kwargs):
    """ Read evtx files or directories
    Input:
        evtxf: evtx filename/directory.
    Return:
        return the filename as xml string.
    """     

    # Parse Arguments
    tool                     = kwargs.get('tool',                    'evtx_parser')
    hostname                 = kwargs.get('hostname',                '')
    do_harmonize             = kwargs.get('harmonize',               True)
    build_all_df             = kwargs.get('build_all_df'            , False)
    plugin                   = 'windows_evtx_record'
    
    # Identifies if input is file or directory and calls the corresponding function

    # Input possibilities:
    # - Single evtx file
    # - Folder with multiple evtx files for one host
    # - Folder with subfolders for multiple hosts,
    #   each of them containing one (or more) evtx files

    pathtype = ""
    if os.path.isdir(path):
        pathtype = "d"
    elif os.path.isfile(path):
        pathtype = "f"
    
    dfs = {}

    if pathtype == "f":
        dfs = read_evtx_file(path,verbose, tool=tool, hostname=hostname, plugin=plugin, orchestrator=None)
    elif pathtype == "d":
        dfs = read_evtx_directory(path,verbose, tool=tool, hostname=hostname, plugin=plugin, orchestrator=None)
    else:
        raise Exception('The File or Path not exist.')
    
    # Harmonize - - - - - - - - - - - - - - - - - - - - - - - - - -
    if do_harmonize :
        print("- Harmonizing pandas dataframes: ")
        
        orchestrator = dfs['D4_Orchestrator_'].iloc[0]        
        tool         = dfs['D4_Tool_'].iloc[0]        
        plugin       = dfs['D4_Plugin_'].iloc[0]        
        hostname     = dfs['D4_Hostname_'].iloc[0]        
        
        if pathtype == "f":
            dfs = harmonize(dfs, tool=tool, plugin="windows_evtx_record", hostname=hostname, orchestrator=orchestrator)
        elif pathtype == "d":
            keys = dfs.keys()
            for key in keys:
                print('  + %-45s ... ' % (key), end='')
                if  len(dfs[key]) > 0:
                    dfs[key] = harmonize(dfs[key], tool=tool, plugin=key, hostname=hostname, orchestrator=orchestrator)
                    print(' [OK]')
                else:
                    print(' [EMPTY]')

                if build_all_df :
                    dfs['all'] = pd.concat([dfs['all'], dfs[key]], ignore_index=True)

        print("- Done Harmonizing.\n")

    
    return dfs

def read_evtx_file_fast(evtxf, verbose=False, **kwargs):
    # This function requires the evtx_dump executable 
    # https://github.com/omerbenamram/evtx

    import re
    import os
    import subprocess
    from evtx import PyEvtxParser

    orchestrator = kwargs.get('orchestrator', None)
    tool         = kwargs.get('tool',         None)
    plugin       = kwargs.get('plugin',       None)
    hostname     = kwargs.get('hostname',     None)

    evtxfbase = re.sub('^.*\\\\', '', evtxf)
    print("- Processing: " + evtxfbase)

    print('- Executing evtx_dump')
    result = subprocess.run(['evtx_dump', evtxf], stdout=subprocess.PIPE)
    xmlstr = result.stdout.decode('ascii')
    xmlstr = re.sub('^Record .*','',xmlstr)
    xmlstr = re.sub('.*xml\ version.*','',xmlstr)
    xmlstr = '<Events>' + xmlstr + '</Events>'
    xmlstr = '<?xml version="1.1" encoding="utf-8" standalone="yes" ?>' + xmlstr

    # print("- Fixing xml via BeautifulSoup")
    # from bs4 import BeautifulSoup
    # xmlstr = BeautifulSoup(xmlstr, 'xml')
    # print("  + xmlstr type: ",end='')
    # print(type(xmlstr))

    #     # Read & Print file size
    #     fsize = os.path.getsize(evtxf)
    #     if verbose :
    #         print("- Size: "+str(fsize))

    # Convert from XML to DF
    convertmethod = 2

    print('- Converting XML to DF')

    if convertmethod == 1:
        # TODO: FIX: This should be much faster but doesn't work
        print("- Fixing XML with lxml")
        from lxml import etree
        root = etree.fromstring(xmlstr.encode('utf-8'))
        xmlstrfixed = etree.tostring(root)

        print("- Converting XML to DF with pd.read_xml")
        #from io import StringIO
        #df = pd.read_xml(StringIO(xmlstr))
        df = pd.read_xml(xmlstr)
    else:
        df = d4utl.xml_to_df(xmlstr,sep=" > ").d4evtx.column_types_set()

    datatype     = "evtx-raw"
    # Insert Tool / Data related HAM Columns
    if not 'D4_Hostname_' in df.columns:
        df.insert(0, 'D4_Hostname_', hostname)
    if not 'D4_Plugin_' in df.columns:
        df.insert(0, 'D4_Plugin_', plugin)
    if not 'D4_Tool_' in df.columns:
        df.insert(0, 'D4_Tool_', tool)
    if not 'D4_Orchestrator_' in df.columns:
        df.insert(0, 'D4_Orchestrator_', orchestrator)
    if not 'D4_DataType_' in df.columns:
        df.insert(0, 'D4_DataType_', datatype)

    if not 'evtxFileName_' in df.columns:
        df.insert(0, 'evtxFileName_', evtxfbase)

    dfs = d4evtx.evtid_dfs_build(df)

    return dfs

def read_evtx_file(evtxf, verbose=False, **kwargs):
    """ Read evtx file.
    Input:
        evtxf: evtx filename        
        verbose: True by default.
    Return:
        return xml string.
    """   
    
    orchestrator = kwargs.get('orchestrator', None)
    tool         = kwargs.get('tool',         None)
    plugin       = kwargs.get('plugin',       None)
    hostname     = kwargs.get('hostname',     None)
      
    # Read & Print file size
    fsize = os.path.getsize(evtxf)
    if verbose :
        print("- Size: "+str(fsize))
    
    print("- Processing:")
    evtxfbase = re.sub('^.*\\\\', '', evtxf)
    print("  + "+evtxfbase)
        
    # Read file (evtx or xml)
    _filename, file_extension = os.path.splitext(evtxf)
    if file_extension == ".evtx":
        xmlstr = evtx_xml(evtxf)
    elif file_extension == ".xml" or file_extension == ".XML":
        with open(evtxf, 'r') as file:
            xmlstr = file.read()
    else:
        print("ERROR: Unsupported file extension / format")
        return

    if xmlstr == "":
        print("ERROR: No XML was read")
        return 

    dfs = {}
        
    try:        
        dfs = d4utl.xml_to_df(xmlstr,sep=" > ").d4evtx.column_types_set() 
    except:
        dfs = pd.DataFrame()          
            
    datatype     = "evtx-raw"        
    # Insert Tool / Data related HAM Columns
    if not 'D4_Hostname_' in dfs.columns:
        dfs.insert(0, 'D4_Hostname_', hostname)
    if not 'D4_Plugin_' in dfs.columns:
        dfs.insert(0, 'D4_Plugin_', plugin)
    if not 'D4_Tool_' in dfs.columns:
        dfs.insert(0, 'D4_Tool_', tool)
    if not 'D4_Orchestrator_' in dfs.columns:         
        dfs.insert(0, 'D4_Orchestrator_', orchestrator)
    if not 'D4_DataType_' in dfs.columns:
        dfs.insert(0, 'D4_DataType_', datatype)
    
    if not 'evtxFileName_' in dfs.columns:
        dfs.insert(0, 'evtxFileName_', evtxfbase)
    
    return dfs

def read_evtx_directory(path,verbose=True, **kwargs):
    """ Read directory of evtx files.
    Input:
        path: evtx filename/directory.
        verbose: True by default.
    Return:
        return the directory filenames as one xml string.
    """      

    dfs={}
    
    print("- Processing:")
    
    # See if we have 2 levels
    files = glob.glob(path+"/*/*.evtx") 
    nl2files=len(files)
    if nl2files != 0:
        # We seem to have 2 levels (dir/hostdirs/*.evtx)
        for file in files:
            # We'll derive the hostname from the dirname
            dirname = os.path.dirname(file)
            dirnamebase = os.path.basename(dirname)
            hostname=dirnamebase
            print("- Reading "+file)
            #dfs=read_evtx_file(file,hostname)
            dfs[hostname]=read_evtx_file(file,hostname, **kwargs)
            evtxfbase=re.sub('^.*\\\\','',file)
            dfs[hostname]['evtxFileName_']=evtxfbase
    else:
        # We seem to have only one level (hostdir/*.evtx)
        files = glob.glob(path+"/*.evtx") 
        nl1files=len(files)
        if nl1files != 0:
            for file in files:
                dirname = os.path.dirname(file)
                dirnamebase = os.path.basename(dirname)
                print("- Reading "+file)
                evtxfbase = re.sub('^.*\\\\','',file)
                dfs[evtxfbase] = read_evtx_file(file, **kwargs)
                dfs['evtxFileName_'] = evtxfbase
        else:
            print("ERROR: Could find any evtx file")

    return dfs


# HARMONIZATION FUNCTIONS #####################################################

def harmonize(df, **kwargs):

    # Specific Harmonization Pre-Processing ----------------------------------- 

    # Name index as pevtnum (plaso evt number) and reset it (as a std col)
#    df.index = df.index.set_names("pevtnum")
#    df = df.reset_index()

    # Resort columns
#    cols = df.columns.tolist()
#    cols.insert(len(cols), cols.pop(cols.index('message')))
#    cols.insert(len(cols), cols.pop(cols.index('pevtnum')))
#    cols.insert(0, cols.pop(cols.index('timestamp_desc')))
#    df = df.reindex(columns=cols)

#    df['timestamp_desc'] = df['timestamp_desc'].str.replace(' Time$','').str.replace('Modification','Modif.')

    # Generic Harmonization ---------------------------------------------------
    #df = d4com.harmonize_common(df, **kwargs)

    # Specific Harmonization Post-Processing ----------------------------------

    # Convert timestamp to datetime
#    nowus = int(time.time()*1000000)
#    df.insert(5, 'Timestamp_', None)
#    df['Timestamp_'] = df['timestamp']
#    df['Timestamp_'] = df['Timestamp_'].mask(df['Timestamp_'].lt(0),0)
#    df['Timestamp_'] = df['Timestamp_'].mask(df['Timestamp_'].gt(nowus),0)
#    df['Timestamp_'] = pd.to_datetime(df['Timestamp_'], unit="us")

    # Resort columns
    #cols = df.columns.tolist()

    # return ------------------------------------------------------------------
    return df

# ANALYSIS FUNCTIONS ##########################################################

# simple ======================================================================
def simple_func(df, *args, **kwargs):
    """ Reformat the input df so the data is presented to the analyst in the
        friendliest possible way

    Parameters:
    df  (pd.dataframe):  Input data 
    
    Returns:
    pd.DataFrame: Optionally it will return the filtered dataframe, 
                  only if ret=True is set, constant & hidden columns included
                  If ret_out=True is set, then the output just as it is shown
                  (without constant/hidden columns) will be return
    """

    # Artifact-specific argument parsing =======================================

    # Variables ================================================================
    hiddencols =  []

    dfout = df

    # Maximum number of lines in DF for beautification
    maxdfbprintlines = 20

    # Pre-Processing ==========================================================

    # Call to simple_common ===================================================
    dfout = d4com.simple_common(df, *args, **kwargs, hiddencols=hiddencols, maxdfbprintlines=maxdfbprintlines)

    # Post-Processing =========================================================

    # Return ==================================================================
    return dfout

# analysis ====================================================================
def analysis(obj, *args, **kwargs):
    """ Redirects execution to analysis_func()
    """
    return analysis_func(obj, *args, **kwargs)

def analysis_func(obj, *args, **kwargs):
    """ Umbrella function that redirects to different types of analysis 
        available on the input data

    Parameters:
    obj:          Input data (typically DF or dict of DFs)
    
    Returns:
    pd.DataFrame: Refer to each specific analysis function
    """

    def syntax():
        print('Syntax: analysis(obj, "analysis_type")\n')
        d4list("str-help")
        return

    def d4list(objtype):

        # Analysis Modules Available for this objective
        anlav = False
        print("Available XXXXX analysis types:")
        if objtype == None or objtype == "str-help" or objtype == "str-list" or  re.match("^dict-pandas_dataframe-XXXXX", objtype):
            anlav = True
            print("- XXXXX_files:  No.events XXXXX file (Input: XXXdfs)")

        if anlav == False:
            print('- No analysis modules available for this object ('+objtype+').')

    nargs = len(args)

    if nargs == 0:
        syntax()
        return

    obj = args[0]

    objtype = d4com.data_identify(obj)

    if isinstance(obj, str):
        if obj == "list":
            d4list(objtype)
            return
        if obj == "help":
            syntax()
            return

    if nargs == 1:
        syntax()
        return

    anltype = args[1]

    if not isinstance(anltype, str):
        syntax()
        return

    if anltype == "help":
        syntax()
        return
    elif anltype == "list":
        d4list(objtype)
        return

    # ANALYSIS FUNCTIONS ======================================================

    # XXXdfs ------------------------------------------------------------------
    # if   re.match("^dict-pandas_dataframe-XXXXX", objtype):
    #     if anltype == "XXXXX_files":
    #         return analysis_XXXXX_files(*args, **kwargs)

    print("INFO: [d4evtx] No analysis functions available for this data type ("+objtype+")")

# DATAFRAME ACCESSOR ##########################################################

@pd.api.extensions.register_dataframe_accessor("d4evtx_parser")
@pd.api.extensions.register_dataframe_accessor("d4_evtx_parser")
class Ds4n6EvtxParserAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def simple(self, *args, **kwargs):
        """ Redirects execution to simple_func()
        """
        df = self._obj
        return simple_func(df, *args, **kwargs)




# DS4N6
#
# Description: library of functions to appy Data Science in several forensics
#              artifacts
#

###############################################################################
# INFO
###############################################################################
# Recommended "import as": d4utl

###############################################################################
# IMPORTS
###############################################################################

# python IMPORTS --------------------------------------------------------------
import os
import re
import inspect
import xmltodict
import glob

# DS IMPORTS ------------------------------------------------------------------
import numpy  as np
import pandas as pd
import ipywidgets as ui
from IPython.display import display, Markdown, HTML

import ds4n6_lib.d4          as d4
import ds4n6_lib.common  as d4com


###############################################################################
# FUNCTIONS
###############################################################################

# EXTENDER pandas COLUMN/ROW FUNCTIONS ########################################
def insert_column_after_another_by_name(*args, **kwargs):
    return insert_or_move_column_after_another_by_name(*args, **kwargs)

def move_column_after_another_by_name(*args, **kwargs):
    return insert_or_move_column_after_another_by_name(*args, **kwargs)

def insert_or_move_column_after_another_by_name(df, col, refcol, defvalue=None, where2move="after"):
    """
    This function supports both inserting a new column or moving a column to
    the designated location
    """
    # Locate index of refcol
    loc = df.columns.get_loc(refcol)

    # New location
    if where2move == "after":
       newloc = loc + 1
    elif where2move == "before":
       newloc = loc - 1

    df.insert(loc=newloc, column=col+'_%%NEW%%_', value=defvalue) 

    # Insert col at location
    if col in df.columns:
        df[col+'_%%NEW%%_'] = df[col]
        df.drop(columns=col, inplace=True)

    df = df.rename(columns={col+'_%%NEW%%_': col})

    return df

# FILE READING/SAVING FUNCTIONS ###############################################
def find_files_in_folder(folder, filter="", ext='', maxdepth=4):

    def filter_files(objs):
        # Check which objects are files
        files = []
        for obj in objs:
            if re.search(filter, obj):
                if os.path.isfile(obj):
                    files.append(obj)

        return files

    if d4.debug >= 4:
        print("DEBUG: - Max Depth:   "+str(maxdepth))
        print("DEBUG: - Top Folder:  "+folder)

    lev       = 1
    files     = []
    nnewfiles = 0
    nfiles    = 0
    wildcard  ="/*"

    while lev <= maxdepth:
        path_name = "{}{}{}".format(folder, wildcard*lev, ext)
        newfiles = filter_files(glob.glob(path_name))
        if newfiles == []:
            nnewfiles = 0
        else:
            files.extend(newfiles)
            nnewfiles = len(newfiles)
            nfiles    = len(files)

        if d4.debug >= 4:
            print("DEBUG: - Finding objs at level {}".format(lev))
            print("DEBUG:     "+path_name)
            print("DEBUG:   + "+str(nnewfiles)+" new   files found.")
            print("DEBUG:   + "+str(nfiles)+" total files found.")

        lev += 1

    if d4.debug >= 4:
        print("DEBUG: - Total Files Found:   "+str(nfiles))
    
    return files

# Read/Save from/to HDF =======================================================

def save_df_to_hdf_chunked(df, filename, chunksize=1000000, complib='blosc', complevel=9):
    print("Saving DF to HDF Store")
    print("")
    print("- Details:")
    print("  + File Name:  "+filename)
    print("  + DF Length:  "+str(len(df)))
    print("  + Chunk Size: "+str(chunksize))
    print("  + No. Chunks: "+str(int(len(df)/chunksize)))
    print("  + Comp. Lib.: "+complib)
    print("  + Comp. Lev.: "+str(complevel))
    print("")

    print("- Saving", end='')
    start = 0
    end = chunksize-1
    cnt = 1
    while end < df.shape[0]:         
        chunk = df.iloc[start:end]
        #chunk.to_hdf(filename, 'df'+str(int(cnt)), mode='a', format='table')
        chunk.to_hdf(filename, 'df'+str(cnt), mode='a', format='table', complib=complib, complevel=complevel)
        start += chunksize
        end += chunksize
        cnt += 1
        
        if cnt % 10 == 0:
            print("["+str(cnt)+"]", end='')
        else:
            print(".", end='')

    print("")
    print("- Done")

def read_df_from_hdf_chunked(filename, chunksize=1000000):

    # Figure out the no. of chunks
    with pd.HDFStore(filename) as hdf:
        import re
        pat = re.compile(r'^/df[0-9][0-9]*$')
        dfkeys = pd.Series([i for i in hdf.keys() if pat.match(i)])
        dfkeys = dfkeys.str.replace('^/df','').astype(int)
        maxdfidx = max(dfkeys)

    print("Details:")
    print("  + File Name:  "+filename)
    print("  + Chunk Size: "+str(chunksize))
    print("  + No. Chunks: "+str(maxdfidx))
    print("")

    print("- Reading", end='')
    with pd.HDFStore(filename, mode='r') as hdf:
        cnt = 1
        df = pd.DataFrame()
        while cnt <= maxdfidx:
            chunkdf = hdf.get('df'+str(cnt))
            df = pd.concat([df, chunkdf], ignore_index=True)
            cnt += 1

            if cnt % 10 == 0:
                print('['+str(cnt)+']', end='')
            else:
                print('.', end='')
    print("")
    print("- Resulting DF:")
    print("  + DF Length:  "+str(len(df))) 
    print("")    
    
    return df

# DISPLAY FUNCTIONS ###########################################################
def display_side_by_side(dfs:list, captions:list):
    """Display tables side by side to save vertical space
    Input:
        dfs: list of pandas.DataFrame
        captions: list of table captions
    """
    output = ""
    combined = dict(zip(captions, dfs))
    for caption, df in combined.items():
        output += df.style.set_table_attributes("style='display:inline'").set_caption(caption)._repr_html_()
        output += "\xa0\xa0\xa0"
    display(HTML(output))


# dict FUNCTIONS ##############################################################
def dfsdict_stats(dfsdict):
    """ Show stats of len keys of dictionary
    Input:
        dfsdict: dictionary        
    """
    for key in dfsdict.keys():
        print('%8i - %-45s' % (len(dfsdict[key]),key))


def dict_to_flatdict(dct,sep="_"):
    """Flaten the dictionary 
    Input:
        dct: dictionary
        sep: separator
    """
    import collections

    flatdict = collections.OrderedDict()

    def recurse(t,parent_key=""):

        if isinstance(t,list):
            for i in range(len(t)):
                recurse(t[i],parent_key + sep + str(i) if parent_key else str(i))
        elif isinstance(t,dict):
            for k,v in t.items():
                recurse(v,parent_key + sep + k if parent_key else k)
        else:
            obj[parent_key] = t
    
    for _l1k,l1v in dct['Events'].items():
        for x in range(len(l1v)):
            obj = collections.OrderedDict()
            recurse(l1v[x])
            flatdict[x]=obj            

    return flatdict


def dict_to_df(dct,sep="_"):
    """Convert dictionary to dataframe
    Input:
        dct: dictionary
        sep: separator
    """
    flatdict = dict_to_flatdict(dct, sep=sep)

    # Let's clear dct to free memory
    dct = {}

    df = pd.DataFrame.from_dict(flatdict, orient='index')

    return df


# XML  FUNCTIONS ##############################################################
def xml_to_df(xmlstr, sep="_"):
    """Convert xml to dataframe
    Input:
        xmlstr: string (xml content)
        sep: separator
    """
    import xmltodict
    # unicode escape invalid XML strings
    xmlstr = escapeInvalidXML(xmlstr)

    try:
        xmldct = xmltodict.parse(xmlstr)
    # except Exception as err:
    except xmltodict.expat.ExpatError as err:
        print("    => ERROR: line: {} column: {} error_string: {}".format(err.lineno,err.offset, xmlstr.splitlines()[err.lineno-1]))
        raise

    # Let's clear xmlstr to free memory. We no longer need it.
    xmlstr=""

    return dict_to_df(xmldct, sep=sep)


def escapeInvalidXML(string):
    """ Escape invalid XML from ranges invalid unicode chars """

    r = re.compile('[^\x09\x0A\x0D\x20-\x7E\x85\xA0-\xFF' \
      + '\u0100-\uD7FF\uE000-\uFDCF\uFDE0-\uFFFD]')

    def replacer(m):
        return "<u>\\u"+('%04X' % ord(m.group(0)))+"</u>"
  
    try:
        return re.sub(r,replacer,string)
    except:
        return string

# df HELPERS ##################################################################
def collapse_constant_columns(df):
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    dfnrows = len(df)

    if dfnrows != 0:
        concolsdf = pd.DataFrame()

        ncols  = df.shape[1]
        colcnt = 1
        for col in df.columns:
            if d4.debug >= 4:
                print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()] collapse_constant_cols -> ["+str(colcnt)+"/"+str(ncols)+"] "+col)
 
            if col != 'nan':
                if df[col].dtype == object:
                    duplicates = df[col].astype('string').drop_duplicates()
                else:
                    duplicates = df[col].drop_duplicates()
                if len(duplicates) == 1:
                    concolsdf = concolsdf.append([[col,str(duplicates.iloc[0])]], ignore_index=True)
                    df = df.drop(columns=col)

            colcnt +=1

        concolsdf = concolsdf.rename(columns={ 0: 'Column', 1: 'Value' })

        return concolsdf, df

# NOTEBOOK FUNCTIONS ##########################################################

def nbgrep(nb,regex,celltype="code"):
    """
    Input:
        nb: notebook file name.
        regex: regex 
        celltype: celltype to apply regex (code by default)
    """
    import nbformat
    import re

    if nb == "":
        print("- ERROR: notebook file name empty.")
        return

    if os.path.exists(nb):
        allhits = []
        nb = nbformat.read(nb,as_version=4)
        for c in dict(nb)['cells']:
            cdict = dict(c)
            if cdict['cell_type'] == celltype:
                cellsrc = cdict['source']
                hits = re.findall(regex,cellsrc)
                allhits = allhits+hits

        return allhits
    else:
        print("- ERROR: notebook file not found.")
        print("         "+nb)
        return


def extract_dfs_from_type(df_type, dfs):
    new_dfs = {}
    for dfkey in dfs:
        if d4com.data_identify(dfs[dfkey]).startswith(df_type):
            new_dfs[dfkey] = dfs[dfkey]
    return new_dfs

# ACCESSOR ####################################################################

@pd.api.extensions.register_dataframe_accessor("d4utl")
class Ds4n6UtilsAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def dummy(self,arg1):
        dummy=0

class PathSelector():

    def __init__(self,start_dir,select_file=True):
        self.file        = None
        self.select_file = select_file
        self.cwd         = start_dir
        self.select      = ui.SelectMultiple(options=['init'],value=(),rows=10,description='')
        self.accord      = ui.Accordion(children=[self.select])

        self.accord.selected_index = None # Start closed (showing path only)
        self.refresh(self.cwd)
        self.select.observe(self.on_update,'value')

    def on_update(self,change):
        if len(change['new']) > 0:
            self.refresh(change['new'][0])

    def refresh(self,item):
        global path
        path = os.path.abspath(os.path.join(self.cwd,item))

        if os.path.isfile(path):
            if self.select_file:
                self.accord.set_title(0,path)
                self.file = path
                self.accord.selected_index = None
            else:
                self.select.value = ()

        else: # os.path.isdir(path)
            self.file = None
            self.cwd  = path

            # Build list of files and dirs
            keys = ['[..]']
            for item in os.listdir(path):
                if item[0] == '.':
                    continue
                elif os.path.isdir(os.path.join(path,item)):
                    keys.append('['+item+']')
                else:
                    keys.append(item)

            # Sort and create list of output values
            keys.sort(key=str.lower)
            vals = []
            for k in keys:
                if k[0] == '[':
                    vals.append(k[1:-1]) # strip off brackets
                else:
                    vals.append(k)

            # Update widget
            self.accord.set_title(0,path)
            self.select.options = list(zip(keys,vals))
            with self.select.hold_trait_notifications():
                self.select.value = ()

def df_outlier_analysis(indf,sensitivity):

    # Method: analyze all fields and find the ones different from others
    # Select cols that have a small no of different values
    # - df.nunique() < 10          --> Bool Series w/ cols that have < 10 different values
    # - df.T[df.nunique() < 10].T  --> Show those values
    # - drop_duplicates()          --> Drop duplicate rows
    # - .index                     --> Show row # for those rows
    # - .iloc                      --> Select df rows based on row index

    # Readable version
    #      intcolsmsk=df.nunique() < 10
    #      inting=df.iloc[df.T[intcolsmsk].T.drop_duplicates().index]
    #      return inting
    # One liner version

    df = indf.copy()
    nrows = df.shape[0]
    maxrows = int(sensitivity*(nrows*.02))
    df.drop(columns=df.columns[df.nunique()>maxrows], inplace=True) 
    intgdf = df.iloc[df.T[df.nunique() < maxrows/5].T.drop_duplicates().index]

    return intgdf


# KNOWLEDGE ###################################################################

# Regex Patterns ==============================================================
ipregex=r"^(?:[0-9]{1,3}\.){3}[0-9]{1,3}$"




###############################################################################
# KNOWLEDGE
###############################################################################

###############################################################################
# IMPORTS
###############################################################################
# DS IMPORTS ------------------------------------------------------------------
import pandas as pd


###############################################################################
# VARIABLES
###############################################################################

# PROCESSES ###################################################################
critical_processes = [
    'System', 'smss.exe', 'wininit.exe', 'RuntimeBroker.exe', 'taskhostw.exe', 'winlogon.exe',
    'csrss.exe', 'services.exe', 'svchost.exe', 'lsaiso.exe', 'lsass.exe', 'explorer.exe']

boot_start_processes = [
    'System', 'smss.exe', 'wininit.exe', 'winlogon.exe', 'csrss.exe', 'services.exe', 'lsaiso.exe',
    'lsass.exe' ]

process_parents = pd.DataFrame([
    ['System', ''],
    ['smss.exe', 'System'],
    ['wininit.exe', 'smss.exe'],
    ['RuntimeBroker.exe', 'svchost.exe'],
    ['taskhostw.exe', 'svchost.exe'],
    ['winlogon.exe', 'smss.exe'],
    ['csrss.exe', 'smss.exe'],
    ['services.exe', 'wininit.exe'],
    ['svchost.exe', 'services.exe'],
    ['lsaiso.exe', 'wininit.exe'],
    ['lsass.exe', 'wininit.exe'],
    ['explorer.exe', 'userinit.exe']],
    columns=['Child', 'Parent'])


# kansa -----------------------------------------------------------------------
kansakeys = ['Arp', 'TempDirListing', 'WMIEvtConsumer', 'SvcAll', 'SmbSession', 'Autorunsc', 'PrefetchListing', 'SigcheckSystemRoot', 'LocalAdmins', 'WMIEvtFilter', 'WMIFltConBind', 'SvcFail', 'LogUserAssist', 'DNSCache', 'Netstat', 'ProcsWMI', 'PSProfiles', 'HostInfo', 'AppCompatCache', 'Tasklistv', 'SvcTrigs']

# Tool DF Columns =============================================================

# volatility ------------------------------------------------------------------
tool_df_columns = {
    'volatility': {
        'apihooks':        { 'HookMode', 'HookType', 'Process', 'PID', 'VictimModule', 'VictimModBase', 'VictimModSize', 'Function', 'HookAddress', 'HookingModule', 'DataAddress', 'Data'},
        'atoms':           { 'Offset(V)', 'Session', 'WindowStation', 'Atom', 'RefCount', 'HIndex', 'Pinned', 'Name'},
        'atomscan':        { 'Offset(P)', 'AtomOfs(V)', 'Atom', 'Refs', 'Pinned', 'Name'},
        'callbacks':       { 'Type', 'Callback', 'Module', 'Details'},
        'devicetree':      { 'Offset(P)', 'Pointers', 'Handles', 'Start', 'Size', 'Service Key', 'Name', 'Driver Name'},
        'dlldump':         { 'Process(V)', 'Name', 'Module Base', 'Module Name', 'Result'},
        'dlllist':         { 'Pid', 'Base', 'Size', 'LoadCount', 'LoadTime', 'Path'},
        'driverirp':       { 'Offset(P)', 'Pointers', 'Handles', 'Start', 'Size', 'Service Key', 'Name', 'Driver Name'},
        'driverscan':      { 'Offset(P)', 'Pointers', 'Handles', 'Start', 'Size', 'Service Key', 'Name', 'Driver Name'},
        'envars':          { 'Pid', 'Process', 'Block', 'Variable', 'Value'},
        'getservicesids':  { 'SID', 'Service'},
        'getsids':         { 'PID', 'Process', 'SID', 'Name'},
        'hashdump':        { 'Hash'},
        'hivelist':        { 'Virtual', 'Physical', 'Name'},
        'hivescan':        { 'Offset(P)'},
        'imageinfo':       { 'Suggested Profile(s)', 'AS Layer1', 'AS Layer2', 'PAE type', 'DTB', 'KDBG', 'Number of Processors', 'Image Type (Service Pack)', 'KPCR for CPU 0', 'KPCR for CPU 1', 'KUSER_SHARED_DATA', 'Image date and time', 'Image local date and time'},
        'ldrmodules':      { 'Pid', 'Process', 'Base', 'InLoad', 'InInit', 'InMem', 'MappedPath'},
        'malfind':         { 'Process', 'Pid', 'Address', 'VadTag', 'Protection', 'Flags', 'Data'},
        'mbrparser':       { 'Offset', 'DiskSignature', 'BootMD5', 'FullBootMD5', 'Distance', 'PartABootFlag', 'PartAType', 'PartALBA', 'PartAStartCHS', 'PartAEndCHS', 'PartASectorSize', 'PartBBootFlag', 'PartBType', 'PartBLBA', 'PartBStartCHS', 'PartBEndCHS', 'PartBSectorSize', 'PartCBootFlag', 'PartCType', 'PartCLBA', 'PartCStartCHS', 'PartCEndCHS', 'PartCSectorSize', 'PartDBootFlag', 'PartDType', 'PartDLBA', 'PartDStartCHS', 'PartDEndCHS', 'PartDSectorSize', 'Bootcode'},
        'memmap':          { 'Process', 'PID', 'Virtual', 'Physical', 'Size', 'DumpFileOffset'},
        'mftparser':       { 'Scanning for MFT entries and building directory, this can take a while'},
        'moddump':         { 'Module Base', 'Module Name', 'Result'},
        'modscan':         { 'Offset(P)', 'Name', 'Base', 'Size', 'File'},
        'modules':         { 'Offset(V)', 'Name', 'Base', 'Size', 'File'},
        'mutantscan':      { 'Offset(P)', 'Pointers', 'Handles', 'Signal', 'Thread', 'CID', 'Name'},
        'privs':           { 'Pid', 'Process', 'Value', 'Privilege', 'Attributes', 'Description'},
        'pslist':          { 'Offset(V)', 'Name', 'PID', 'PPID', 'Thds', 'Hnds', 'Sess', 'Wow64', 'Start', 'Exit'},
        'psscan':          { 'Offset(P)', 'Name', 'PID', 'PPID', 'PDB', 'Time Created', 'Time Exited'},
        'pstree':          { 'Offset', 'Name', 'Pid', 'PPid', 'Thds', 'Hnds', 'Time'},
        'psxview':         { 'Offset(P)', 'Name', 'PID', 'pslist', 'psscan', 'thrdproc', 'pspcid', 'csrss', 'session', 'deskthrd', 'ExitTime'},
        'ssdt':            { 'Table', 'TableOffset', 'NumEntries', 'Entry', 'Addr', 'Function', 'Owner', 'Destination', 'HookName'},
        'symlinkscan':     { 'Offset(P)', 'Pointers', 'Handles', 'Creation Time', 'Origin', 'Target'},
        'thrdscan':        { 'Offset(P)', 'PID', 'TID', 'Start Address', 'Create Time', 'Exit Time'},
        'threads':         { 'Offset', 'PID', 'TID', 'Tags', 'Create Time', 'Exit Time', 'Owning Process', 'Attached Process', 'State', 'State Reason', 'Base Priority', 'Priority', 'TEB', 'Start Address', 'Owner Name', 'Win32 Start Address', 'Win32 Thread', 'Cross Thread Flags', 'EIP', 'EAX', 'EBX', 'ECX', 'EDX', 'ESI', 'EDI', 'ESP', 'EBP', 'ErrCode', 'SegCS', 'SegSS', 'SegDS', 'SegES', 'SegGS', 'SegFS', 'EFlags', 'dr0', 'dr1', 'dr2', 'dr3', 'dr6', 'dr7', 'SSDT', 'Entry Number', 'Descriptor Service Table', 'Hook Number', 'Function Name', 'Function Address', 'Module Name', 'Disassembly'},
        'timeliner':       { 'Start', 'Header', 'Item', 'Details'},
        'timers':          { 'Offset(V)', 'DueTime', 'Period(ms)', 'Signaled', 'Routine', 'Module'},
        'unloadedmodules': { 'Name', 'StartAddress', 'EndAddress', 'Time'}
    }
}


# EVENT IDs ###################################################################
evtid_desc = { 
    'Security.evtx': { 
        1100: 'The event logging service has shut down',
        1101: 'Audit events have been dropped by the transport.',
        1102: 'The audit log was cleared',
        1104: 'The security Log is now full',
        1105: 'Event log automatic backup',
        1108: 'The event logging service encountered an error',
        4608: 'Windows is starting up',
        4609: 'Windows is shutting down',
        4610: 'An authentication package has been loaded by the Local Security Authority',
        4611: 'A trusted logon process has been registered with the Local Security Authority',
        4612: 'Internal resources allocated for the queuing of audit messages have been exhausted, leading to the loss of some audits.',
        4614: 'A notification package has been loaded by the Security Account Manager.',
        4615: 'Invalid use of LPC port',
        4616: 'The system time was changed.',
        4618: 'A monitored security event pattern has occurred',
        4621: 'Administrator recovered system from CrashOnAuditFail',
        4622: 'A security package has been loaded by the Local Security Authority.',
        4624: 'An account was successfully logged on',
        4625: 'An account failed to log on',
        4626: 'User/Device claims information',
        4627: 'Group membership information.',
        4634: 'An account was logged off',
        4646: 'IKE DoS-prevention mode started',
        4647: 'User initiated logoff',
        4648: 'A logon was attempted using explicit credentials',
        4649: 'A replay attack was detected',
        4650: 'An IPsec Main Mode security association was established',
        4651: 'An IPsec Main Mode security association was established',
        4652: 'An IPsec Main Mode negotiation failed',
        4653: 'An IPsec Main Mode negotiation failed',
        4654: 'An IPsec Quick Mode negotiation failed',
        4655: 'An IPsec Main Mode security association ended',
        4656: 'A handle to an object was requested',
        4657: 'A registry value was modified',
        4658: 'The handle to an object was closed',
        4659: 'A handle to an object was requested with intent to delete',
        4660: 'An object was deleted',
        4661: 'A handle to an object was requested',
        4662: 'An operation was performed on an object',
        4663: 'An attempt was made to access an object',
        4664: 'An attempt was made to create a hard link',
        4665: 'An attempt was made to create an application client context.',
        4666: 'An application attempted an operation',
        4667: 'An application client context was deleted',
        4668: 'An application was initialized',
        4670: 'Permissions on an object were changed',
        4671: 'An application attempted to access a blocked ordinal through the TBS',
        4672: 'Special privileges assigned to new logon',
        4673: 'A privileged service was called',
        4674: 'An operation was attempted on a privileged object',
        4675: 'SIDs were filtered',
        4688: 'A new process has been created',
        4689: 'A process has exited',
        4690: 'An attempt was made to duplicate a handle to an object',
        4691: 'Indirect access to an object was requested',
        4692: 'Backup of data protection master key was attempted',
        4693: 'Recovery of data protection master key was attempted',
        4694: 'Protection of auditable protected data was attempted',
        4695: 'Unprotection of auditable protected data was attempted',
        4696: 'A primary token was assigned to process',
        4697: 'A service was installed in the system',
        4698: 'A scheduled task was created',
        4699: 'A scheduled task was deleted',
        4700: 'A scheduled task was enabled',
        4701: 'A scheduled task was disabled',
        4702: 'A scheduled task was updated',
        4703: 'A token right was adjusted',
        4704: 'A user right was assigned',
        4705: 'A user right was removed',
        4706: 'A new trust was created to a domain',
        4707: 'A trust to a domain was removed',
        4709: 'IPsec Services was started',
        4710: 'IPsec Services was disabled',
        4711: 'PAStore Engine (1%)',
        4712: 'IPsec Services encountered a potentially serious failure',
        4713: 'Kerberos policy was changed',
        4714: 'Encrypted data recovery policy was changed',
        4715: 'The audit policy (SACL) on an object was changed',
        4716: 'Trusted domain information was modified',
        4717: 'System security access was granted to an account',
        4718: 'System security access was removed from an account',
        4719: 'System audit policy was changed',
        4720: 'A user account was created',
        4722: 'A user account was enabled',
        4723: 'An attempt was made to change an account\'s password',
        4724: 'An attempt was made to reset an accounts password',
        4725: 'A user account was disabled',
        4726: 'A user account was deleted',
        4727: 'A security-enabled global group was created',
        4728: 'A member was added to a security-enabled global group',
        4729: 'A member was removed from a security-enabled global group',
        4730: 'A security-enabled global group was deleted',
        4731: 'A security-enabled local group was created',
        4732: 'A member was added to a security-enabled local group',
        4733: 'A member was removed from a security-enabled local group',
        4734: 'A security-enabled local group was deleted',
        4735: 'A security-enabled local group was changed',
        4737: 'A security-enabled global group was changed',
        4738: 'A user account was changed',
        4739: 'Domain Policy was changed',
        4740: 'A user account was locked out',
        4741: 'A computer account was created',
        4742: 'A computer account was changed',
        4743: 'A computer account was deleted',
        4744: 'A security-disabled local group was created',
        4745: 'A security-disabled local group was changed',
        4746: 'A member was added to a security-disabled local group',
        4747: 'A member was removed from a security-disabled local group',
        4748: 'A security-disabled local group was deleted',
        4749: 'A security-disabled global group was created',
        4750: 'A security-disabled global group was changed',
        4751: 'A member was added to a security-disabled global group',
        4752: 'A member was removed from a security-disabled global group',
        4753: 'A security-disabled global group was deleted',
        4754: 'A security-enabled universal group was created',
        4755: 'A security-enabled universal group was changed',
        4756: 'A member was added to a security-enabled universal group',
        4757: 'A member was removed from a security-enabled universal group',
        4758: 'A security-enabled universal group was deleted',
        4759: 'A security-disabled universal group was created',
        4760: 'A security-disabled universal group was changed',
        4761: 'A member was added to a security-disabled universal group',
        4762: 'A member was removed from a security-disabled universal group',
        4763: 'A security-disabled universal group was deleted',
        4764: 'A groups type was changed',
        4765: 'SID History was added to an account',
        4766: 'An attempt to add SID History to an account failed',
        4767: 'A user account was unlocked',
        4768: 'A Kerberos authentication ticket (TGT) was requested',
        4769: 'A Kerberos service ticket was requested',
        4770: 'A Kerberos service ticket was renewed',
        4771: 'Kerberos pre-authentication failed',
        4772: 'A Kerberos authentication ticket request failed',
        4773: 'A Kerberos service ticket request failed',
        4774: 'An account was mapped for logon',
        4775: 'An account could not be mapped for logon',
        4776: 'The domain controller attempted to validate the credentials for an account',
        4777: 'The domain controller failed to validate the credentials for an account',
        4778: 'A session was reconnected to a Window Station',
        4779: 'A session was disconnected from a Window Station',
        4780: 'The ACL was set on accounts which are members of administrators groups',
        4781: 'The name of an account was changed',
        4782: 'The password hash an account was accessed',
        4783: 'A basic application group was created',
        4784: 'A basic application group was changed',
        4785: 'A member was added to a basic application group',
        4786: 'A member was removed from a basic application group',
        4787: 'A non-member was added to a basic application group',
        4788: 'A non-member was removed from a basic application group..',
        4789: 'A basic application group was deleted',
        4790: 'An LDAP query group was created',
        4791: 'A basic application group was changed',
        4792: 'An LDAP query group was deleted',
        4793: 'The Password Policy Checking API was called',
        4794: 'An attempt was made to set the Directory Services Restore Mode administrator password',
        4797: 'An attempt was made to query the existence of a blank password for an account',
        4798: 'A user\'s local group membership was enumerated.',
        4799: 'A security-enabled local group membership was enumerated',
        4800: 'The workstation was locked',
        4801: 'The workstation was unlocked',
        4802: 'The screen saver was invoked',
        4803: 'The screen saver was dismissed',
        4816: 'RPC detected an integrity violation while decrypting an incoming message',
        4817: 'Auditing settings on object were changed.',
        4818: 'Proposed Central Access Policy does not grant the same access permissions as the current Central Access Policy',
        4819: 'Central Access Policies on the machine have been changed',
        4820: 'A Kerberos Ticket-granting-ticket (TGT) was denied because the device does not meet the access control restrictions',
        4821: 'A Kerberos service ticket was denied because the user, device, or both does not meet the access control restrictions',
        4822: 'NTLM authentication failed because the account was a member of the Protected User group',
        4823: 'NTLM authentication failed because access control restrictions are required',
        4824: 'Kerberos preauthentication by using DES or RC4 failed because the account was a member of the Protected User group',
        4825: 'A user was denied the access to Remote Desktop. By default, users are allowed to connect only if they are members of the Remote Desktop Users group or Administrators group',
        4826: 'Boot Configuration Data loaded',
        4830: 'SID History was removed from an account',
        4864: 'A namespace collision was detected',
        4865: 'A trusted forest information entry was added',
        4866: 'A trusted forest information entry was removed',
        4867: 'A trusted forest information entry was modified',
        4868: 'The certificate manager denied a pending certificate request',
        4869: 'Certificate Services received a resubmitted certificate request',
        4870: 'Certificate Services revoked a certificate',
        4871: 'Certificate Services received a request to publish the certificate revocation list (CRL)',
        4872: 'Certificate Services published the certificate revocation list (CRL)',
        4873: 'A certificate request extension changed',
        4874: 'One or more certificate request attributes changed.',
        4875: 'Certificate Services received a request to shut down',
        4876: 'Certificate Services backup started',
        4877: 'Certificate Services backup completed',
        4878: 'Certificate Services restore started',
        4879: 'Certificate Services restore completed',
        4880: 'Certificate Services started',
        4881: 'Certificate Services stopped',
        4882: 'The security permissions for Certificate Services changed',
        4883: 'Certificate Services retrieved an archived key',
        4884: 'Certificate Services imported a certificate into its database',
        4885: 'The audit filter for Certificate Services changed',
        4886: 'Certificate Services received a certificate request',
        4887: 'Certificate Services approved a certificate request and issued a certificate',
        4888: 'Certificate Services denied a certificate request',
        4889: 'Certificate Services set the status of a certificate request to pending',
        4890: 'The certificate manager settings for Certificate Services changed.',
        4891: 'A configuration entry changed in Certificate Services',
        4892: 'A property of Certificate Services changed',
        4893: 'Certificate Services archived a key',
        4894: 'Certificate Services imported and archived a key',
        4895: 'Certificate Services published the CA certificate to Active Directory Domain Services',
        4896: 'One or more rows have been deleted from the certificate database',
        4897: 'Role separation enabled',
        4898: 'Certificate Services loaded a template',
        4899: 'A Certificate Services template was updated',
        4900: 'Certificate Services template security was updated',
        4902: 'The Per-user audit policy table was created',
        4904: 'An attempt was made to register a security event source',
        4905: 'An attempt was made to unregister a security event source',
        4906: 'The CrashOnAuditFail value has changed',
        4907: 'Auditing settings on object were changed',
        4908: 'Special Groups Logon table modified',
        4909: 'The local policy settings for the TBS were changed',
        4910: 'The group policy settings for the TBS were changed',
        4911: 'Resource attributes of the object were changed',
        4912: 'Per User Audit Policy was changed',
        4913: 'Central Access Policy on the object was changed',
        4928: 'An Active Directory replica source naming context was established',
        4929: 'An Active Directory replica source naming context was removed',
        4930: 'An Active Directory replica source naming context was modified',
        4931: 'An Active Directory replica destination naming context was modified',
        4932: 'Synchronization of a replica of an Active Directory naming context has begun',
        4933: 'Synchronization of a replica of an Active Directory naming context has ended',
        4934: 'Attributes of an Active Directory object were replicated',
        4935: 'Replication failure begins',
        4936: 'Replication failure ends',
        4937: 'A lingering object was removed from a replica',
        4944: 'The following policy was active when the Windows Firewall started',
        4945: 'A rule was listed when the Windows Firewall started',
        4946: 'A change has been made to Windows Firewall exception list. A rule was added',
        4947: 'A change has been made to Windows Firewall exception list. A rule was modified',
        4948: 'A change has been made to Windows Firewall exception list. A rule was deleted',
        4949: 'Windows Firewall settings were restored to the default values',
        4950: 'A Windows Firewall setting has changed',
        4951: 'A rule has been ignored because its major version number was not recognized by Windows Firewall',
        4952: 'Parts of a rule have been ignored because its minor version number was not recognized by Windows Firewall',
        4953: 'A rule has been ignored by Windows Firewall because it could not parse the rule',
        4954: 'Windows Firewall Group Policy settings has changed. The new settings have been applied',
        4956: 'Windows Firewall has changed the active profile',
        4957: 'Windows Firewall did not apply the following rule',
        4958: 'Windows Firewall did not apply the following rule because the rule referred to items not configured on this computer',
        4960: 'IPsec dropped an inbound packet that failed an integrity check',
        4961: 'IPsec dropped an inbound packet that failed a replay check',
        4962: 'IPsec dropped an inbound packet that failed a replay check',
        4963: 'IPsec dropped an inbound clear text packet that should have been secured',
        4964: 'Special groups have been assigned to a new logon',
        4965: 'IPsec received a packet from a remote computer with an incorrect Security Parameter Index (SPI).',
        4976: 'During Main Mode negotiation, IPsec received an invalid negotiation packet.',
        4977: 'During Quick Mode negotiation, IPsec received an invalid negotiation packet.',
        4978: 'During Extended Mode negotiation, IPsec received an invalid negotiation packet.',
        4979: 'IPsec Main Mode and Extended Mode security associations were established.',
        4980: 'IPsec Main Mode and Extended Mode security associations were established',
        4981: 'IPsec Main Mode and Extended Mode security associations were established',
        4982: 'IPsec Main Mode and Extended Mode security associations were established',
        4983: 'An IPsec Extended Mode negotiation failed',
        4984: 'An IPsec Extended Mode negotiation failed',
        4985: 'The state of a transaction has changed',
        5024: 'The Windows Firewall Service has started successfully',
        5025: 'The Windows Firewall Service has been stopped',
        5027: 'The Windows Firewall Service was unable to retrieve the security policy from the local storage',
        5028: 'The Windows Firewall Service was unable to parse the new security policy.',
        5029: 'The Windows Firewall Service failed to initialize the driver',
        5030: 'The Windows Firewall Service failed to start',
        5031: 'The Windows Firewall Service blocked an application from accepting incoming connections on the network.',
        5032: 'Windows Firewall was unable to notify the user that it blocked an application from accepting incoming connections on the network',
        5033: 'The Windows Firewall Driver has started successfully',
        5034: 'The Windows Firewall Driver has been stopped',
        5035: 'The Windows Firewall Driver failed to start',
        5037: 'The Windows Firewall Driver detected critical runtime error. Terminating',
        5038: 'Code integrity determined that the image hash of a file is not valid',
        5039: 'A registry key was virtualized.',
        5040: 'A change has been made to IPsec settings. An Authentication Set was added.',
        5041: 'A change has been made to IPsec settings. An Authentication Set was modified',
        5042: 'A change has been made to IPsec settings. An Authentication Set was deleted',
        5043: 'A change has been made to IPsec settings. A Connection Security Rule was added',
        5044: 'A change has been made to IPsec settings. A Connection Security Rule was modified',
        5045: 'A change has been made to IPsec settings. A Connection Security Rule was deleted',
        5046: 'A change has been made to IPsec settings. A Crypto Set was added',
        5047: 'A change has been made to IPsec settings. A Crypto Set was modified',
        5048: 'A change has been made to IPsec settings. A Crypto Set was deleted',
        5049: 'An IPsec Security Association was deleted',
        5050: 'An attempt to programmatically disable the Windows Firewall using a call to INetFwProfile.FirewallEnabled(FALSE',
        5051: 'A file was virtualized',
        5056: 'A cryptographic self test was performed',
        5057: 'A cryptographic primitive operation failed',
        5058: 'Key file operation',
        5059: 'Key migration operation',
        5060: 'Verification operation failed',
        5061: 'Cryptographic operation',
        5062: 'A kernel-mode cryptographic self test was performed',
        5063: 'A cryptographic provider operation was attempted',
        5064: 'A cryptographic context operation was attempted',
        5065: 'A cryptographic context modification was attempted',
        5066: 'A cryptographic function operation was attempted',
        5067: 'A cryptographic function modification was attempted',
        5068: 'A cryptographic function provider operation was attempted',
        5069: 'A cryptographic function property operation was attempted',
        5070: 'A cryptographic function property operation was attempted',
        5071: 'Key access denied by Microsoft key distribution service',
        5120: 'OCSP Responder Service Started',
        5121: 'OCSP Responder Service Stopped',
        5122: 'A Configuration entry changed in the OCSP Responder Service',
        5123: 'A configuration entry changed in the OCSP Responder Service',
        5124: 'A security setting was updated on OCSP Responder Service',
        5125: 'A request was submitted to OCSP Responder Service',
        5126: 'Signing Certificate was automatically updated by the OCSP Responder Service',
        5127: 'The OCSP Revocation Provider successfully updated the revocation information',
        5136: 'A directory service object was modified',
        5137: 'A directory service object was created',
        5138: 'A directory service object was undeleted',
        5139: 'A directory service object was moved',
        5140: 'A network share object was accessed',
        5141: 'A directory service object was deleted',
        5142: 'A network share object was added.',
        5143: 'A network share object was modified',
        5144: 'A network share object was deleted.',
        5145: 'A network share object was checked to see whether client can be granted desired access',
        5146: 'The Windows Filtering Platform has blocked a packet',
        5147: 'A more restrictive Windows Filtering Platform filter has blocked a packet',
        5148: 'The Windows Filtering Platform has detected a DoS attack and entered a defensive mode; packets associated with this attack will be discarded.',
        5149: 'The DoS attack has subsided and normal processing is being resumed.',
        5150: 'The Windows Filtering Platform has blocked a packet.',
        5151: 'A more restrictive Windows Filtering Platform filter has blocked a packet.',
        5152: 'The Windows Filtering Platform blocked a packet',
        5153: 'A more restrictive Windows Filtering Platform filter has blocked a packet',
        5154: 'The Windows Filtering Platform has permitted an application or service to listen on a port for incoming connections',
        5155: 'The Windows Filtering Platform has blocked an application or service from listening on a port for incoming connections',
        5156: 'The Windows Filtering Platform has allowed a connection',
        5157: 'The Windows Filtering Platform has blocked a connection',
        5158: 'The Windows Filtering Platform has permitted a bind to a local port',
        5159: 'The Windows Filtering Platform has blocked a bind to a local port',
        5168: 'Spn check for SMB/SMB2 fails.',
        5169: 'A directory service object was modified',
        5170: 'A directory service object was modified during a background cleanup task',
        5376: 'Credential Manager credentials were backed up',
        5377: 'Credential Manager credentials were restored from a backup',
        5378: 'The requested credentials delegation was disallowed by policy',
        5379: 'Credential Manager credentials were read',
        5380: 'Vault Find Credential',
        5381: 'Vault credentials were read',
        5382: 'Vault credentials were read',
        5440: 'The following callout was present when the Windows Filtering Platform Base Filtering Engine started',
        5441: 'The following filter was present when the Windows Filtering Platform Base Filtering Engine started',
        5442: 'The following provider was present when the Windows Filtering Platform Base Filtering Engine started',
        5443: 'The following provider context was present when the Windows Filtering Platform Base Filtering Engine started',
        5444: 'The following sub-layer was present when the Windows Filtering Platform Base Filtering Engine started',
        5446: 'A Windows Filtering Platform callout has been changed',
        5447: 'A Windows Filtering Platform filter has been changed',
        5448: 'A Windows Filtering Platform provider has been changed',
        5449: 'A Windows Filtering Platform provider context has been changed',
        5450: 'A Windows Filtering Platform sub-layer has been changed',
        5451: 'An IPsec Quick Mode security association was established',
        5452: 'An IPsec Quick Mode security association ended',
        5453: 'An IPsec negotiation with a remote computer failed because the IKE and AuthIP IPsec Keying Modules (IKEEXT) service is not started',
        5456: 'PAStore Engine applied Active Directory storage IPsec policy on the computer',
        5457: 'PAStore Engine failed to apply Active Directory storage IPsec policy on the computer',
        5458: 'PAStore Engine applied locally cached copy of Active Directory storage IPsec policy on the computer',
        5459: 'PAStore Engine failed to apply locally cached copy of Active Directory storage IPsec policy on the computer',
        5460: 'PAStore Engine applied local registry storage IPsec policy on the computer',
        5461: 'PAStore Engine failed to apply local registry storage IPsec policy on the computer',
        5462: 'PAStore Engine failed to apply some rules of the active IPsec policy on the computer',
        5463: 'PAStore Engine polled for changes to the active IPsec policy and detected no changes',
        5464: 'PAStore Engine polled for changes to the active IPsec policy, detected changes, and applied them to IPsec Services',
        5465: 'PAStore Engine received a control for forced reloading of IPsec policy and processed the control successfully',
        5466: 'PAStore Engine polled for changes to the Active Directory IPsec policy, determined that Active Directory cannot be reached, and will use the cached copy of the Active Directory IPsec policy instead',
        5467: 'PAStore Engine polled for changes to the Active Directory IPsec policy, determined that Active Directory can be reached, and found no changes to the policy',
        5468: 'PAStore Engine polled for changes to the Active Directory IPsec policy, determined that Active Directory can be reached, found changes to the policy, and applied those changes',
        5471: 'PAStore Engine loaded local storage IPsec policy on the computer',
        5472: 'PAStore Engine failed to load local storage IPsec policy on the computer',
        5473: 'PAStore Engine loaded directory storage IPsec policy on the computer',
        5474: 'PAStore Engine failed to load directory storage IPsec policy on the computer',
        5477: 'PAStore Engine failed to add quick mode filter',
        5478: 'IPsec Services has started successfully',
        5479: 'IPsec Services has been shut down successfully',
        5480: 'IPsec Services failed to get the complete list of network interfaces on the computer',
        5483: 'IPsec Services failed to initialize RPC server. IPsec Services could not be started',
        5484: 'IPsec Services has experienced a critical failure and has been shut down',
        5485: 'IPsec Services failed to process some IPsec filters on a plug-and-play event for network interfaces',
        5632: 'A request was made to authenticate to a wireless network',
        5633: 'A request was made to authenticate to a wired network',
        5712: 'A Remote Procedure Call (RPC) was attempted',
        5888: 'An object in the COM+ Catalog was modified',
        5889: 'An object was deleted from the COM+ Catalog',
        5890: 'An object was added to the COM+ Catalog',
        6144: 'Security policy in the group policy objects has been applied successfully',
        6145: 'One or more errors occured while processing security policy in the group policy objects',
        6272: 'Network Policy Server granted access to a user',
        6273: 'Network Policy Server denied access to a user',
        6274: 'Network Policy Server discarded the request for a user',
        6275: 'Network Policy Server discarded the accounting request for a user',
        6276: 'Network Policy Server quarantined a user',
        6277: 'Network Policy Server granted access to a user but put it on probation because the host did not meet the defined health policy',
        6278: 'Network Policy Server granted full access to a user because the host met the defined health policy',
        6279: 'Network Policy Server locked the user account due to repeated failed authentication attempts',
        6280: 'Network Policy Server unlocked the user account',
        6281: 'Code Integrity determined that the page hashes of an image file are not valid...',
        6400: 'BranchCache: Received an incorrectly formatted response while discovering availability of content.',
        6401: 'BranchCache: Received invalid data from a peer. Data discarded.',
        6402: 'BranchCache: The message to the hosted cache offering it data is incorrectly formatted.',
        6403: 'BranchCache: The hosted cache sent an incorrectly formatted response to the client\'s message to offer it data.',
        6404: 'BranchCache: Hosted cache could not be authenticated using the provisioned SSL certificate.',
        6405: 'BranchCache: %2 instance(s) of event id %1 occurred.',
        6406: '%1 registered to Windows Firewall to control filtering for the following:',
        6407: '%1',
        6408: 'Registered product %1 failed and Windows Firewall is now controlling the filtering for %2.',
        6409: 'BranchCache: A service connection point object could not be parsed',
        6410: 'Code integrity determined that a file does not meet the security requirements to load into a process. This could be due to the use of shared sections or other issues',
        6416: 'A new external device was recognized by the system.',
        6417: 'The FIPS mode crypto selftests succeeded',
        6418: 'The FIPS mode crypto selftests failed',
        6419: 'A request was made to disable a device',
        6420: 'A device was disabled',
        6421: 'A request was made to enable a device',
        6422: 'A device was enabled',
        6423: 'The installation of this device is forbidden by system policy',
        6424: 'The installation of this device was allowed, after having previously been forbidden by policy',
        8191: 'Highest System-Defined Audit Message Value',
    },

    'Microsoft-Windows-TaskScheduler%4Operational.evtx': {
        101: 'Task Scheduler failed to start the "%1" task for user "%2". The error value is: %3.',
        103: 'Task Scheduler failed to start the "%2" instance of the "%1" task for user "%3". The error value is: %4.',
        104: 'Task Scheduler failed to log on "%1". The failure occurred in "%2". Ensure the credentials for the task are correctly specified. The error value is: %3.',
        106: 'The user "%2" registered the Task Scheduler task "%1".',
        107: 'Task Scheduler launched the "%2" instance of task "%1" due to a time trigger.',
        108: 'Task Scheduler launched the "%2" instance of task "%1". The task was triggered by an event trigger.',
        109: 'Task Scheduler launched "%2" instance of task "%1" due to a registration trigger.',
        110: 'Task Scheduler launched the "%2" instance of task "%1" for user "%3" .',
        112: 'Task Scheduler could not start task "%1" because the network was unavailable. Ensure the computer is connected to the required network as specified in the task. If the task does not require network presence, remove the network condition from the task configuration.',
        114: 'Task Scheduler could not launch task "%1" as scheduled. Instance "%2" is started now as required by the configuration option to start the task when available, if the scheduled time is missed.',
        117: 'Task Scheduler launched the "%2" instance of task "%1" due to an idle condition.',
        118: 'Task Scheduler launched the "%2" instance of task "%1" due to system startup.',
        119: 'Task Scheduler launched the "%3" instance of task "%1" due to user "%2" logon.',
        120: 'Task Scheduler launched the "%3" instance of task "%1" due to user "%2" connecting to the console.',
        121: 'Task Scheduler launched the "%3" instance of task "%1" due to user "%2" disconnecting from the console.',
        122: 'Task Scheduler launched the "%3" instance of task "%1" due to user "%2" remotely connecting.',
        123: 'Task Scheduler launched the "%3" instance of task "%1" due to user "%2" remotely disconnecting.',
        124: 'Task Scheduler launched the "%3" instance of task "%1" due to user "%2" locking the computer.',
        125: 'Task Scheduler launched the "%3" instance of task "%1" due to user "%2" unlocking the computer.',
        126: 'Task Scheduler failed to execute the "%1" task. Task Scheduler is attempting to restart the task. The error value is: %2.',
        127: 'Task Scheduler failed to execute task "%1" due to a shutdown race condition. Task Scheduler is attempting to restart the task.',
        128: 'Task Scheduler did not launch task "%1" because the current time exceeds the configured task end time. Extend the end time boundary for the task if required.',
        130: 'The Task Scheduler service failed to start the "%1" task due to the service being busy. The error value is: %2.',
        140: 'User "%2" updated Task Scheduler task "%1"',
        141: 'User "%2" deleted Task Scheduler task "%1"',
        142: 'User "%2" disabled Task Scheduler task "%1"',
        145: 'Task Scheduler woke up the computer to run a task.',
        200: 'Task Scheduler launched the "%2" action in the "%3" instance of task "%1".',
        201: 'Task Scheduler successfully completed task "%1" , instance "%3" , and action "%2".',
        204: 'Task Scheduler failed to retrieve the event triggering values for task "%1" . The event will be ignored. Additional Data: Error Value: %2.',
        205: 'Task Scheduler failed to match the pattern of events for the following task: "%1" . The events will be ignored. The error value is: %2.',
        305: 'Task Scheduler did not send the "%1" task to task engine "%2" . The error value is: %3.',
        322: 'Task Scheduler did not launch task "%1" because instance "%2" of the same task is already running.',
        324: 'Task Scheduler queued the "%2" instance of task "%1" and will launch it as soon as instance "%3" completes.',
        326: 'Task Scheduler did not launch task "%1" because the computer is running on batteries. If launching the task on batteries is required, change the respective flag in the task configuration.',
        332: 'Task Scheduler did not launch the "%1" task because user "%2" was not logged on when the launching conditions were met. Ensure the user is logged on or change the task definition to allow the task to launch when the user is logged off.'
    },

    'System.evtx': {
         104: 'The audit log was cleared.',
        6005: 'The Event log service was started.',
        6006: 'The Event log service was stopped.'
    }
}

# KNOWLEDGE ENRICHMENT (_K_) ====================================================
evtid_col_map = {
    'Security.evtx': {
        4624: {
            'LogonType':{
                2: 'Interactive',
                3: 'Network',
                4: 'Batch',
                5: 'Service',
                7: 'Unlock',
                8: 'NetworkCleartext',
                9: 'NewCredentials',
                10: 'RemoteInteractive',
                11: 'CachedInteractive',
                12: 'CachedRemoteInteractive',
                13: 'CachedUnlock'
             }
        },
        4625: {
            'LogonType':{
                2: 'Interactive',
                3: 'Network',
                4: 'Batch',
                5: 'Service',
                7: 'Unlock',
                8: 'NetworkCleartext',
                9: 'NewCredentials',
                10: 'RemoteInteractive',
                11: 'CachedInteractive',
                12: 'CachedRemoteInteractive',
                13: 'CachedUnlock'
             },
            'ErrorCode':{
                '0xC0000064': 'Invalid Username',
                '0xC000006A': 'Incorrect Password',
                '0xC000006F': 'Account Time Restriction',
                '0xC0000070': 'Account Source Restriction',
                '0xC0000071': 'Password Expired',
                '0xC0000072': 'Account Disabled',
                '0xC0000193': 'Account Expired',
                '0xC0000234': 'Account Locked'
             }
        },
        4697: {
            'ServiceStartType': {
                0: 'Boot',
                1: 'System',
                2: 'Autoload',
                3: 'Manual',
                4: 'Disabled'
             }
        },
        4771: {
            'ErrorCode':{
                '0x6': 'Invalid Username',
                '0x7': 'Server Not Found',
                '0xC': 'Policy Restriction',
                '0x12': 'Account Disabled',
                '0x17': 'Password Expired',
                '0x18': 'Password Invalide',
                '0x25': 'Excessive Clock Skew'
             }
        },
        4776: {
            'ErrorCode':{
                '0xC0000064': 'Invalid Username',
                '0xC000006A': 'Incorrect Password',
                '0xC000006F': 'Account Time Restriction',
                '0xC0000070': 'Account Source Restriction',
                '0xC0000071': 'Password Expired',
                '0xC0000072': 'Account Disabled',
                '0xC0000193': 'Account Expired',
                '0xC0000234': 'Account Locked'
             }
        }
    }
}

evtid_col_map_desc = {
    'Security.evtx': {
        4697: {
            'ServiceStartType': {
                0: 'Boot - The kernel loaded will load this driver first as its needed to use the boot volume device',
                1: 'System - This is loaded by the I/O subsystem',
                2: 'Autoload - The service is always loaded and run',
                3: 'Manual - This service does not start automatically and must be manually started by the user',
                4: 'Disabled - The service is disabled and should not be started'
             }
        }
    }          
}

# column ENRICHMENT (transforms) ##############################################
#col_transforms = {
#    'Microsoft-Windows-WMI-Activity%4Operational.evtx': {
#        5861: {
#            'UserData > Operation_ESStoConsumerBinding > PossibleCause': ''

# simple FILTERS ##############################################################
evtx_simple_filters = {
    'Security.evtx': {
        4648: {
            'exclude_builtin_accounts': r"not TargetUserName.str.contains(r'^DWM-[0-9][0-9]*$|^UMFD-[0-9][0-9]*$|^SYSTEM$|^NETWORK\ SERVICE$|^LOCAL\ SERVICE$')"
        }
    }          
}

# evtid GROUPS ################################################################
evtid_groups = {
    'Security.evtx': {
        'account_usage': { 4624, 4625, 4634, 4647, 4648, 4672, 4720, 4726 },
        'account_logon': { 4776, 4768, 4769, 4771 },
        'enumeration'  : { 4798, 4799 },
        'shares'       : { 5140, 5141, 5142, 5143, 5144, 5150 },
        'services'     : { 4697 },
        'log_clearing' : { 1102 },
        'runas'        : { 4648 }
    },
    'System.evtx': {
        'services'     : { 7045 },
        'log_clearing' : { 104 },
        'log_service'  : { 6005, 6006 }
    },
    'Microsoft-Windows-WMI-Activity%4Operational.evtx': { 
        'wmi':           { 5857, 5858, 5859, 5860, 5861 } 
    }
}

windows_builtin_accounts_regex = r'^DWM-[0-9][0-9]*$|^UMFD-[0-9][0-9]*$|^SYSTEM$|^NETWORK\ SERVICE$|^LOCAL\ SERVICE$'
wbar = windows_builtin_accounts_regex

# FIND EVIL ###################################################################

# Basic per-event signatures
evilsigs = {
    'Security.evtx': {
        4625: {
            'T1110.001-Brute_Force-Password_Guessing-Attempt': '0 == 0'
        }
    },
    'Microsoft-Windows-WMI-Activity%4Operational.evtx': {
        5861: {
            'T1047-powershell_on_wmi': '`UserData > Operation_ESStoConsumerBinding > PossibleCause`.str.contains("powershell")'
        }
     }
}

# CONFIGURATION ###############################################################
hidden_cols_evtx = ['ProcessID', 'ThreadID', 'CorrelationActivityID', '@CorrelationActivityID']
